from __future__ import annotations

import itertools
from typing import TYPE_CHECKING

import torch

from ..typedefs._defs import Trajectory

if TYPE_CHECKING:
    from ..typedefs._params import ModelParams


def check_trajectories(trajectories: list[Trajectory], c: torch.Tensor | None):
    """Check if trajectories are not empty, well sorted and compatible with censoring.

    Args:
        trajectories (list[Trajectory]): The trajectories.
        c (torch.Tensor | None): The censoring times.

    Raises:
        ValueError: If some trajectory is empty.
        ValueError: If some trajectory is not sorted.
        ValueError: If some trajectory is not compatible with the censoring times.
    """
    if any(len(trajectory) == 0 for trajectory in trajectories):
        raise ValueError("Trajectories must not be empty")
    if any(
        not all(t0 <= t1 for t0, t1 in itertools.pairwise(t for t, _ in trajectory))
        for trajectory in trajectories
    ):
        raise ValueError(
            "Trajectories must be sorted by time, in ascending order. Also ensure "
            "there are no NaN values as this will trigger the check"
        )
    if c is not None and any(
        trajectory[-1][0] > c for trajectory, c in zip(trajectories, c, strict=True)
    ):
        raise ValueError(
            "Last trajectory time transition not be greater than censoring time"
        )


def check_matrix_dim(flat: torch.Tensor, dim: int, method: str):
    """Sets dimensions for matrix.

    Args:
        flat (torch.Tensor): The flat tensor.
        dim (int): The dimension of the matrix.
        method (str): The method to use.

    Raises:
        ValueError: If the number of elements is incompatible with method "full".
        ValueError: If the number of elements is incompatible with method "diag".
        ValueError: If the number of elements is not one and the method is "ball".
        ValueError: If the method is not in ("full", "diag", "ball").
    """
    match method:
        case "full":
            if flat.numel() != (dim * (dim + 1)) // 2:
                raise ValueError(
                    f"{flat.numel()} is incompatible with full matrix of dimension "
                    f"{dim}"
                )
        case "diag":
            if flat.numel() != dim:
                raise ValueError(
                    f"{flat.numel()} is incompatible with diag matrix of dimension "
                    f"{dim}"
                )
        case "ball":
            if flat.numel() != 1:
                raise ValueError(
                    f"Expected 1 element for flat, got {flat.numel()} for matrix of "
                    f" dimension {dim}"
                )
        case _:
            raise ValueError(
                f"Method must be be either 'full', 'diag' or 'ball', got {method}"
            )


def check_params_align(params1: ModelParams, params2: ModelParams):
    """Checks if the named parameters lists have the same length, names and methods.

    Args:
        params1 (ModelParams): The first instance of ModelParams.
        params2 (ModelParams): The second instance of ModelParams.

    Raises:
        ValueError: If the parameters do not have the same names.
        ValueError: If the parameters do not have the same shapes.
    """
    if (
        dict(params1.named_parameters()).keys()
        != dict(params2.named_parameters()).keys()
    ):
        raise ValueError("All parameters must have the same names")
    if any(
        t1.shape != t2.shape
        for t1, t2 in zip(params1.parameters(), params2.parameters(), strict=True)
    ):
        raise ValueError("All parameters must have the same shapes")



from collections.abc import Callable
from typing import Any

import torch

from ..typedefs._data import CompleteModelData
from ..typedefs._params import ModelParams


class MCMCMixin:
    """Mixin for MCMC sampling."""

    params: ModelParams
    n_chains: int
    init_step_size: float
    adapt_rate: float
    target_accept_rate: float

    def _logpdfs_aux_fn(
        self,
        data: CompleteModelData,
        b: torch.Tensor,
    ) -> tuple[torch.Tensor, torch.Tensor]: ...

    def __init__(
        self,
        n_chains: int,
        init_step_size: float,
        adapt_rate: float,
        target_accept_rate: float,
        *args: Any,
        **kwargs: Any,
    ):
        """Initializes the MCMC mixin.

        Args:
            n_chains (int): The number of parallel chains to spawn.
            init_step_size (float): Kernel step in Metropolis-Hastings.
            adapt_rate (float): Adaptation rate for the step_size.
            target_accept_rate (float): Mean acceptance target.
        """
        super().__init__(*args, **kwargs)

        self.n_chains = n_chains
        self.init_step_size = init_step_size
        self.adapt_rate = adapt_rate
        self.target_accept_rate = target_accept_rate

    def _init_mcmc(self, data: CompleteModelData):
        """Initializes the MCMC sampler.

        Args:
            data (CompleteModelData): The complete model data.

        Returns:
            MetropolisHastingsSampler: The initialized MCMC sampler.
        """
        return MetropolisHastingsSampler(
            lambda b: self._logpdfs_aux_fn(data, b),
            torch.zeros(self.n_chains, len(data), self.params.q.dim),
            self.n_chains,
            self.init_step_size,
            self.adapt_rate,
            self.target_accept_rate,
        )


class MetropolisHastingsSampler:
    """A robust Metropolis-Hastings sampler with adaptive step size."""

    logpdfs_aux_fn: Callable[[torch.Tensor], tuple[torch.Tensor, torch.Tensor]]
    n_chains: int
    adapt_rate: int | float
    target_accept_rate: int | float
    b: torch.Tensor
    logpdfs: torch.Tensor
    aux: torch.Tensor
    step_sizes: torch.Tensor

    def __init__(
        self,
        logpdfs_aux_fn: Callable[[torch.Tensor], tuple[torch.Tensor, torch.Tensor]],
        init_b: torch.Tensor,
        n_chains: int,
        init_step_size: int | float,
        adapt_rate: int | float,
        target_accept_rate: int | float,
    ):
        """Initializes the Metropolis-Hastings sampler kernel.

        Args:
            logpdfs_aux_fn (Callable[[torch.Tensor], tuple[torch.Tensor,
                torch.Tensor]]): The log pdfs function with auxiliary data.
            init_b (torch.Tensor): Starting b for the chain.
            n_chains (int): The number of parallel chains to spawn.
            init_step_size (int | float): Kernel step in Metropolis-Hastings.
            adapt_rate (int | float): Adaptation rate for the step_size.
            target_accept_rate (int | float): Mean acceptance target.
        """
        self.logpdfs_aux_fn = torch.no_grad()(logpdfs_aux_fn)
        self.n_chains = n_chains
        self.adapt_rate = adapt_rate
        self.target_accept_rate = target_accept_rate

        # Initialize b
        self.b = init_b.clone()

        # Compute initial log logpdfs
        self.logpdfs, self.psi = self.logpdfs_aux_fn(self.b)

        # Proposal noise initialization
        self._noise = torch.empty_like(self.b)
        self.step_sizes = torch.full((1, self.b.size(-2)), init_step_size)

    def step(self):
        """Performs a single kernel step."""
        # Generate proposal noise
        self._noise.normal_()

        # Get the proposal
        proposed_state = self.b + self._noise * self.step_sizes.unsqueeze(-1)
        proposed_logpdfs, proposed_aux = self.logpdfs_aux_fn(proposed_state)
        logpdf_diff = proposed_logpdfs - self.logpdfs

        # Vectorized acceptance decision
        log_uniform = torch.log(torch.rand_like(logpdf_diff))
        accept_mask = log_uniform < logpdf_diff

        torch.where(accept_mask.unsqueeze(-1), proposed_state, self.b, out=self.b)
        torch.where(accept_mask, proposed_logpdfs, self.logpdfs, out=self.logpdfs)
        torch.where(accept_mask.unsqueeze(-1), proposed_aux, self.psi, out=self.psi)

        # Update step sizes
        mean_accept_mask = accept_mask.to(torch.get_default_dtype()).mean(dim=0)
        adaptation = (mean_accept_mask - self.target_accept_rate) * self.adapt_rate
        self.step_sizes *= torch.exp(adaptation)

    def run(self, n_steps: int):
        """Runs the sampler for a given number of steps.

        Args:
            n_steps (int): The number of steps to run the sampler for.
        """
        for _ in range(n_steps):
            self.step()



from collections.abc import Mapping
from dataclasses import dataclass, field
from typing import Any

import torch
from sklearn.base import BaseEstimator  # type: ignore
from sklearn.utils._param_validation import validate_params  # type: ignore
from sklearn.utils.validation import (  # type: ignore
    assert_all_finite,  # type: ignore
    check_consistent_length,  # type: ignore
)

from ..utils._checks import check_trajectories
from ..utils._surv import build_all_buckets
from ._defs import (
    IndividualEffectsFn,
    LinkFn,
    LogBaseHazardFn,
    RegressionFn,
    Trajectory,
)
from ._params import ModelParams


# Dataclasses
@dataclass
class ModelDesign(BaseEstimator):
    """Class containing model design.

    For all functions, please use broadcasting as much as possible. It is almost always
    possible to broadcast parameters to vectorize efficiently the operations. If you
    copy, beware of a heavy performance hit. If unable, please use vmap.

    Also, note that the function passed to the MCMC sampler will be built using the
    `torch.no_grad()` decorator. If needs be, use `torch.enable_grad()` if one of the
    model design functions always require gradient computation regardless of setting.

    Ensure all functions all well defined on a closed interval and are differentiable
    almost everywhere.

    Attributes:
        individual_effects_fn (IndividualEffectsFn): The individual effects function. It
            must be able to yield 2D or 3D tensors given inputs of `gamma` (population
            parameters), `x` (covariates matrix), and `b` (random effects). Note `b` is
            either 2D or 3D.
        regression_fn (RegressionFn): The regression function. It
            must be able to yield 3D and 4D tensors given 1D or 2D time inputs, as well
            as `psi` input of order 2 or 3. This is not very restrictive, but requires
            to be careful. The last dimension is the dimension of the response variable;
            second last is the repeated measurements; third last is individual based;
            possible fourth last is for parallelization of the MCMC sampler.
        surv_map (Mapping[tuple[Any, Any], tuple[LogBaseHazardFn, LinkFn]]): A mapping
            of transition keys that can be typed however you want. The tuple contains a
            log base hazard function, as well as a link function that shares the same
            requirements as `regression_fn`. Log base hazard function is expected to be
            pure if caching is enabled, otherwise it will lead to false computations.

    Examples:
        >>> def sigmoid(t: torch.Tensor, psi: torch.Tensor):
        ...     scale, offset, slope = psi.chunk(3, dim=-1)
        ...     # Fully broadcasted
        ...     return (scale * torch.sigmoid((t - offset) / slope)).unsqueeze(-1)
        >>> individual_effects_fn = lambda gamma, x, b: gamma + b
        >>> regression_fn = sigmoid
        >>> surv_map = {("alive", "dead"): (Exponential(1.2), sigmoid)}
    """

    individual_effects_fn: IndividualEffectsFn
    regression_fn: RegressionFn
    surv_map: Mapping[
        tuple[Any, Any],
        tuple[LogBaseHazardFn, LinkFn],
    ]


@dataclass
class ModelData(BaseEstimator):
    r"""Dataclass containing learnable multistate joint model data.

    Note `y` is expected to be a 3D tensor of dimension :math:`(n, m, d)` if there are
    :math:`n` individual, with a maximum number of :math:`m` measurements in
    :math:`\mathbb{R}^d`. Its values should not be all NaNs.

    Raises:
        ValueError: If some trajectory is empty.
        ValueError: If some trajectory is not sorted.
        ValueError: If some trajectory is not compatible with the censoring times.
        ValueError: If any of the inputs contain inf or NaN values except `y`.
        ValueError: If the size is not consistent between inputs.

    Attributes:
        x (torch.Tensor | None): The fixed covariates.
        t (torch.Tensor): The measurement times. Either a 1D tensor if the times are
            shared by all individual, or a matrix of individual times. Use padding with
            NaNs when necessary.
        y (torch.Tensor): The measurements. A 3D tensor of dimension :math:`(n, m, d)`
            if there are :math:`n` individual, with a maximum number of :math:`m`
            measurements in :math:`\mathbb{R}^d`. Use padding with NaNs when
            necessary.
        trajectories (list[Trajectory]): The list of the individual trajectories.
            A `Trajectory` is a list of tuples containing time and state.
        c (torch.Tensor): The censoring times as a column vector. They must not
            be less than the trajectory maximum times.
    """

    x: torch.Tensor | None
    t: torch.Tensor
    y: torch.Tensor
    trajectories: list[Trajectory]
    c: torch.Tensor

    def __len__(self) -> int:
        """Gets the number of individuals.

        Returns:
            int: The number of individuals.
        """
        return len(self.trajectories)

    def __post_init__(self):
        """Runs the post init conversions.

        Raises:
            ValueError: If some trajectory is empty.
            ValueError: If some trajectory is not sorted.
            ValueError: If some trajectory is not compatible with the censoring times.
            ValueError: If any of the inputs contain inf or NaN values except `t` and
                `y`.
            ValueError: If the size is not consistent between inputs.
        """
        validate_params(
            {
                "x": [torch.Tensor],
                "t": [torch.Tensor],
                "y": [torch.Tensor],
                "trajectories": [list],
                "c": [torch.Tensor],
                "skip_validation": [bool],
            },
            prefer_skip_nested_validation=True,
        )

        check_trajectories(self.trajectories, self.c)

        assert_all_finite(self.x, input_name="x")
        assert_all_finite(self.t, input_name="t", allow_nan=True)
        assert_all_finite(self.y, input_name="y", allow_nan=True)
        assert_all_finite(self.c, input_name="c")

        check_consistent_length(self.x, self.y, self.c, self.trajectories)
        check_consistent_length(self.t.transpose(0, -1), self.y.transpose(0, -2))

        # Check NaNs between t and y
        if ((~self.y.isnan()).any(dim=-1) & self.t.isnan()).any():
            raise ValueError("NaN time values on non NaN y values")


class CompleteModelData(ModelData):
    """Complete model data class.

    Use this to compute the log pdfs by hand if necessary. It is possible to simply
    call the `MultiStateJointModel` class as it inherits from `nn.Module`.

    Attributes:
        valid_mask (torch.Tensor): The mask of the valid measurements.
        n_valid (torch.Tensor): The number of valid measurements.
        valid_t (torch.Tensor): The valid times.
        valid_y (torch.Tensor): The valid measurements.
        buckets (dict[tuple[Any, Any], tuple[torch.Tensor, ...]]): The buckets for the
            trajectories.
    """

    valid_mask: torch.Tensor = field(init=False)
    n_valid: torch.Tensor = field(init=False)
    valid_t: torch.Tensor = field(init=False)
    valid_y: torch.Tensor = field(init=False)
    buckets: dict[tuple[Any, Any], tuple[torch.Tensor, ...]] = field(init=False)

    @validate_params(
        {
            "model_design": [ModelDesign],
            "params": [ModelParams],
        },
        prefer_skip_nested_validation=True,
    )
    def prepare(self, model_design: ModelDesign, params: ModelParams):
        """Sets the missing representation.

        Args:
            model_design (ModelDesign): The design of the model.
            params (ModelParams): The model parameters.
        """
        check_consistent_length(params.r.cov, self.y.transpose(0, -1))

        self.valid_mask = ~self.y.isnan()
        self.n_valid = self.valid_mask.sum(dim=-2).to(torch.get_default_dtype())
        self.valid_t = self.t.nan_to_num(self.t.nanmean().item())
        self.valid_y = self.y.nan_to_num()
        self.buckets = build_all_buckets(
            self.trajectories, self.c, tuple(model_design.surv_map.keys())
        )


@dataclass
class SampleData(BaseEstimator):
    """Dataclass for data used in sampling.

    Raises:
        ValueError: If some trajectory is empty.
        ValueError: If some trajectory is not sorted.
        ValueError: If some trajectory is not compatible with the censoring times.
        ValueError: If any of the inputs contain inf or NaN values.
        ValueError: If the size is not consistent between inputs.

    Attributes:
        x (torch.Tensor | None): The fixed covariates.
        trajectories (list[Trajectory]): The list of the individual trajectories.
            A `Trajectory` is a list of tuples containing time and state.
        psi (torch.Tensor): The individual parameters. Define it as a matrix with
            the same number of rows as there are `len(trajectories)`. Only use a 3D
            tensor if you fully understand the codebase and the mechanisms. Trajectory
            sampling may only be used with matrices.
        c (torch.Tensor | None, optional): The censoring times as a column vector. They
            must not be less than the trajectory maximum times. This corresponds to
            the last times of observation of the individuals or prediction current
            times.
    """

    x: torch.Tensor | None
    trajectories: list[Trajectory]
    psi: torch.Tensor
    c: torch.Tensor | None = None

    def __len__(self) -> int:
        """Gets the number of individuals.

        Returns:
            int: The number of individuals.
        """
        return len(self.trajectories)

    def __post_init__(self):
        """Runs the post init conversions and checks.

        Raises:
            ValueError: If some trajectory is empty.
            ValueError: If some trajectory is not sorted.
            ValueError: If some trajectory is not compatible with the censoring times.
            ValueError: If any of the inputs contain inf or NaN values.
            ValueError: If the size is not consistent between inputs.
        """
        validate_params(
            {
                "trajectories": [list],
                "psi": [torch.Tensor],
                "c": [torch.Tensor],
            },
            prefer_skip_nested_validation=True,
        )

        check_trajectories(self.trajectories, self.c)

        assert_all_finite(self.x, input_name="x")
        assert_all_finite(self.psi, input_name="psi")
        assert_all_finite(self.c, input_name="c")

        check_consistent_length(
            self.x, self.psi.transpose(0, -2), self.c, self.trajectories
        )



"""Functions for the jmstate package."""

from ._base_hazards import (
    Exponential,
    Gompertz,
    LogNormal,
    Weibull,
    clock_forward,
    clock_reset,
)
from ._individual_effects import gamma_plus_b, gamma_x_plus_b, identity
from ._regression_and_link import linear

__all__ = [
    "Exponential",
    "Gompertz",
    "LogNormal",
    "Weibull",
    "clock_forward",
    "clock_reset",
    "gamma_plus_b",
    "gamma_x_plus_b",
    "identity",
    "linear",
]



import torch


def linear(t: torch.Tensor, psi: torch.Tensor) -> torch.Tensor:
    r"""Implements the linear regression or link function.

    When reverting to a linear joint model, this gives the mapping:

    .. math::
        h(t, \psi) = \psi,

    where :math:`\psi` are the individual parameters.

    Args:
        t (torch.Tensor): The time points.
        psi (torch.Tensor): The individual effects (parameters).

    Returns:
        torch.Tensor: The computed transformation.
    """
    return psi.unsqueeze(-2).expand(*psi.shape[:-1], t.size(-1), -1)



from numbers import Integral, Real
from typing import cast

import torch
from sklearn.base import BaseEstimator  # type: ignore
from sklearn.utils._param_validation import Interval, validate_params  # type: ignore
from sklearn.utils.validation import check_is_fitted  # type: ignore
from torch.nn.utils import parameters_to_vector

from ..typedefs._data import CompleteModelData, ModelDesign
from ..typedefs._params import ModelParams
from ._fit import FitMixin
from ._predict import PredictMixin


class MultiStateJointModel(BaseEstimator, FitMixin, PredictMixin):
    r"""A class of the nonlinear multistate joint model.

    Please note this class encompasses both the linear joint model and the standard
    joint model, but also allows for the modeling of multiple states assuming a semi
    Markov property. The model is defined by a set of longitudinal and hazard
    functions in `ModelDesign`, which are parameterized by a set of parameters in
    `ModelParams`. Parameters may also be shared (see the documentation).

    The model is fit using a stochastic gradient ascent algorithm, and the parameters
    are sampled using a Metropolis-Hastings algorithm.

    Dynamic prediciton is possible using the different prediction methods, that support
    dynamic prediction with both single and double Monte Carlo integration.

    For numerical integration, use `n_quad` to specify the number of nodes for the
    Gauss Legendre quadrature of hazard, and `n_bisect` to specify the number of
    bisection steps for the bisection algorithm for transition sampling.

    For caching, use `cache_limit` to specify the limit of the cache used in hazard
    computation, greatly reducing memory and CPU usage. None means infinite, 0 means
    no caching.

    For MCMC, use `n_chains` to specify the number of chains, `init_step_size` to
    specify the initial step size for the MCMC sampler, `adapt_rate` to specify the
    adaptation rate for the step size, `target_accept_rate` to specify the target
    acceptance rate, `n_warmup` to specify the number of warmup iterations, and
    `n_subsample` to specify the number of subsamples.

    For fitting, use `max_iter_fit` to specify the maximum number of iterations for
    stochastic gradient ascent, `n_samples_summary` to specify the number of samples to
    compute Fisher Information Matrix and model selection criteria, `tol` to specify the
    tolerance for the R2 convergence, and `window_size` to specify the window size for
    the R2 convergence. The longer, the more stable the R2 convergence. The default
    value seems a sweet spot, and the stopping criterion is scale-agnostic. Any
    optimizer can be through the `optimizer` parameter. If set to `None`, then fitting
    is not possible. Recommended to use `torch.optim.Adam` with a learning rate of
    0.1 to 1.0. A value of 0.5 is a good starting point.

    For printing, use `verbose` to specify whether to print the progress of the model
    fitting and predicting.

    Attributes:
        model_design (ModelDesign): The model design.
        params (ModelParams): The (variable) model parameters.
        optimizer (torch.optim.Optimizer | None): The optimizer.
        n_quad (int): The number of nodes for the Gauss Legendre quadrature of hazard.
        n_bisect (int): The number of bisection steps for the bisection algorithm.
        cache_limit (int | None): The limit of the cache used in hazard computation,
            greatly reducing memory and CPU usage. None means infinite, 0 means no
            caching.
        n_chains (int): The number of parallel MCMC chains.
        init_step_size (float): Kernel standard error in Metropolis.
        adapt_rate (float): Adaptation rate for the step_size.
        target_accept_rate (float): Mean acceptance target.
        n_warmup (int): The number of warmup iterations for the MCMC sampler.
        n_subsample (int): The number of subsamples for the MCMC sampler.
        max_iter_fit (int): The maximum number of iterations for stochastic gradient
            ascent.
        tol (float): The tolerance for the R2 convergence.
        window_size (int): The window size for the R2 convergence.
        n_samples_summary (int): The number of samples used to compute Fisher
            Information Matrix and model selection criteria.
        verbose (bool): Whether to print the progress of the model fitting.
        params_vector_history_ (list[torch.Tensor]): The history of model parameters.
        fim_ (torch.Tensor | None): The Fisher Information Matrix.
        loglik_ (float | None): The log likelihood.
        aic_ (float | None): The Akaike Information Criterion.
        bic_ (float | None): The Bayesian Information Criterion.

    Examples:
        >>> # Declares initial model
        >>> optimizer = torch.optim.Adam(init_params.parameters(), lr=0.5)
        >>> model = MultiStateJointModel(model_design, init_params, optimizer)
        >>> # Runs optimization process
        >>> model.fit(data)
        >>> # Prints a summary of the model
        >>> model.summary()
    """

    model_design: ModelDesign
    params: ModelParams
    optimizer: torch.optim.Optimizer | None
    n_quad: int
    n_bisect: int
    cache_limit: int | None
    n_chains: int
    init_step_size: float
    adapt_rate: float
    target_accept_rate: float
    n_warmup: int
    n_subsample: int
    max_iter_fit: int
    tol: float
    window_size: int
    n_samples_summary: int
    verbose: bool
    params_vector_history_: list[torch.Tensor]
    fim_: torch.Tensor | None
    loglik_: float | None
    aic_: float | None
    bic_: float | None

    @validate_params(
        {
            "model_design": [ModelDesign],
            "params": [ModelParams],
            "optimizer": [torch.optim.Optimizer, None],
            "n_quad": [Interval(Integral, 1, None, closed="left")],
            "n_bisect": [Interval(Integral, 1, None, closed="left")],
            "cache_limit": [Interval(Integral, 0, None, closed="left"), None],
            "n_chains": [Interval(Integral, 1, None, closed="left")],
            "init_step_size": [Interval(Real, 0, None, closed="neither")],
            "adapt_rate": [Interval(Real, 0, None, closed="left")],
            "target_accept_rate": [Interval(Real, 0, 1, closed="neither")],
            "n_warmup": [Interval(Integral, 0, None, closed="left")],
            "n_subsample": [Interval(Integral, 0, None, closed="left")],
            "max_iter_fit": [Interval(Integral, 1, None, closed="left")],
            "tol": [Interval(Real, 0, 1, closed="both")],
            "window_size": [Interval(Integral, 1, None, closed="left")],
            "n_samples_summary": [Interval(Integral, 1, None, closed="left")],
            "verbose": [bool],
        },
        prefer_skip_nested_validation=True,
    )
    def __init__(
        self,
        model_design: ModelDesign,
        params: ModelParams,
        optimizer: torch.optim.Optimizer | None = None,
        *,
        n_quad: int = 32,
        n_bisect: int = 32,
        cache_limit: int | None = 256,
        n_chains: int = 5,
        init_step_size: float = 0.1,
        adapt_rate: float = 0.01,
        target_accept_rate: float = 0.234,
        n_warmup: int = 100,
        n_subsample: int = 10,
        max_iter_fit: int = 500,
        tol: float = 0.1,
        window_size: int = 100,
        n_samples_summary: int = 500,
        verbose: bool = True,
    ):
        """Initializes the joint model based on the user defined design.

        Args:
            model_design (ModelDesign): Model design containing modeling information.
            params (ModelParams): (Initial) values for the parameters.
            optimizer (torch.optim.Optimizer | None, optional): The optimizer   used for
                fitting. Defaults to None.
            n_quad (int, optional): The used number of points for Gauss-Legendre
                quadrature. Defaults to 32.
            n_bisect (int, optional): The number of bisection steps used in transition
                sampling. Defaults to 32.
            cache_limit (int | None, optional): The max length of cache. Defaults to
                256.
            n_chains (int, optional): The number of chains for MCMC. Defaults to 5.
            init_step_size (float, optional): The initial step size for the MCMC
                sampler. Defaults to 0.1.
            adapt_rate (float, optional): The adaptation rate for the MCMC sampler.
                Defaults to 0.01.
            target_accept_rate (float, optional): The target acceptance rate for the
                MCMC sampler. Defaults to 0.234.
            n_warmup (int, optional): The number of warmup iterations for the MCMC
                sampler. Defaults to 100.
            n_subsample (int, optional): The number of subsamples for the MCMC sampler.
                Defaults to 10.
            max_iter_fit (int, optional): The maximum number of iterations for
                stochastic gradient ascent. Defaults to 500.
            tol (float, optional): The tolerance for the convergence. Defaults to 0.1.
            window_size (int, optional): The window size for the convergence. Defaults
                to 100.
            n_samples_summary (int, optional): The number of samples used to compute
                Fisher Information Matrix and model selection criteria. Defaults to 500.
            verbose (bool, optional): Whether to print the progress of the model
                fitting. Defaults to True.
        """
        # Info of the Mixin Classes
        super().__init__(
            optimizer=optimizer,
            n_quad=n_quad,
            n_bisect=n_bisect,
            cache_limit=cache_limit,
            n_chains=n_chains,
            init_step_size=init_step_size,
            adapt_rate=adapt_rate,
            target_accept_rate=target_accept_rate,
            max_iter_fit=max_iter_fit,
            tol=tol,
            window_size=window_size,
            n_samples_summary=n_samples_summary,
        )

        # Store model components
        self.model_design = model_design
        self.params = params
        self.n_warmup = n_warmup
        self.n_subsample = n_subsample
        self.verbose = verbose
        self.vector_params_history_ = [
            parameters_to_vector(self.params.parameters()).detach()
        ]
        self.fim_ = None
        self.loglik_ = None
        self.aic_ = None
        self.bic_ = None

    def _logpdfs_aux_fn(
        self, data: CompleteModelData, b: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Gets the log pdfs with individual effects and log likelihoods.

        Args:
            data (CompleteModelData): Dataset on which likelihood is computed.
            b (torch.Tensor): The random effects.

        Returns:
           tuple[torch.Tensor, torch.Tensor]: The log pdfs and aux.
        """
        psi = self.model_design.individual_effects_fn(self.params.gamma, data.x, b)
        logpdfs = (
            super()._long_logliks(data, psi)
            + super()._hazard_logliks(data, psi)
            + super()._prior_logliks(b)
        )

        return logpdfs, psi

    @property
    def stderr(self) -> torch.Tensor:
        r"""Returns the estimated standard error of the parameters as a vector.

        They can be used to draw confidence intervals. The standard errors are computed
        using the diagonal of the inverse of the inverse Fisher Information Matrix at
        the MLE

        .. math::
            \mathrm{stderr} = \sqrt{\operatorname{diag}\left( \hat{\mathcal{I}}_n
            (\hat{\theta})^{-1} \right)}

        Raises:
            ValueError: If the model is not fitted.

        Returns:
            torch.Tensor: The estimated standard error as a vector.
        """
        check_is_fitted(self, "fim_")

        return cast(torch.Tensor, self.fim_).inverse().diag().sqrt()



import torch


def _tril_from_flat(flat: torch.Tensor, dim: int) -> torch.Tensor:
    """Generate the lower triangular matrix associated with flat tensor.

    Args:
        flat (torch.Tensor): Flat tensor
        dim (int): Dimension of the matrix.

    Returns:
        torch.Tensor: The lower triangular matrix.
    """
    return torch.zeros(dim, dim).index_put_(tuple(torch.tril_indices(dim, dim)), flat)


def _flat_from_tril(L: torch.Tensor) -> torch.Tensor:
    """Flatten the lower triangular part (including the diagonal) of a square matrix.

    Into a 1D tensor, in row-wise order.

    Args:
        L (torch.Tensor): Square lower-triangular matrix of shape (dim, dim).

    Raises:
        RuntimeError: If the flattening fails.

    Returns:
        torch.Tensor: Flattened 1D tensor containing the lower triangular entries.
    """
    dim = L.size(0)
    return L[tuple(torch.tril_indices(dim, dim))]


def log_cholesky_from_flat(
    flat: torch.Tensor, dim: int, method: str = "full"
) -> torch.Tensor:
    """Computes log cholesky from flat tensor according to choice of method.

    Args:
        flat (torch.Tensor): The flat tensor parameter.
        dim (int): The dimension of the matrix.
        method (str, optional): The method, full, diag or ball. Defaults to "full".

    Raises:
        ValueError: If the method is not in ("full", "diag", "ball").

    Returns:
        torch.Tensor: The log cholesky representation.
    """
    match method:
        case "full":
            return _tril_from_flat(flat, dim)
        case "diag":
            return torch.diag(flat)
        case "ball":
            return flat * torch.eye(dim)
        case _:
            raise ValueError(
                f"Method must be be either 'full', 'diag' or 'ball', got {method}"
            )


def flat_from_log_cholesky(L: torch.Tensor, method: str = "full") -> torch.Tensor:
    """Computes flat tensor from log cholesky matrix according to choice of method.

    Args:
        L (torch.Tensor): The square lower triangular matrix parameter.
        method (str, optional): The method, full, diag or ball. Defaults to "full".

    Raises:
        ValueError: If the method is not in ("full", "diag", "ball").

    Returns:
        torch.Tensor: The flat representation.
    """
    match method:
        case "full":
            return _flat_from_tril(L)
        case "diag":
            return L.diag()
        case "ball":
            return L[0, 0].reshape(1)
        case _:
            raise ValueError(
                f"Method must be be either 'full', 'diag' or 'ball', got {method}"
            )



from math import ceil
from numbers import Integral
from typing import Any, cast

import torch
from sklearn.utils._param_validation import Interval, validate_params  # type: ignore
from sklearn.utils.validation import (  # type: ignore
    assert_all_finite,  # type: ignore
    check_consistent_length,  # type: ignore
    check_is_fitted,  # type: ignore
)
from torch.nn.utils import parameters_to_vector, vector_to_parameters
from tqdm import trange

from ..typedefs._data import CompleteModelData, ModelData, ModelDesign, SampleData
from ..typedefs._defs import Trajectory
from ..typedefs._params import ModelParams
from ..utils._cache import Cache
from ._hazard import HazardMixin
from ._sampler import MCMCMixin


class PredictMixin(HazardMixin, MCMCMixin):
    """Mixin class for prediction."""

    model_design: ModelDesign
    params: ModelParams
    n_chains: int
    n_warmup: int
    n_subsample: int
    verbose: bool
    fim_: torch.Tensor | None
    _cache: Cache

    def __init__(self, *args: Any, **kwargs: Any):
        """Initialize the prediction mixin."""
        super().__init__(*args, **kwargs)

    def _sample_vector_params(self, sample_size: int) -> torch.Tensor:
        """Sample parameters based on asymptotic behavior of the MLE.

        This uses Bernstein-von Mises theorem to approximate the posterior distribution
        of the parameters as a multivariate normal distribution with mean equal to the
        MLE and covariance matrix equal to the inverse of the Fisher Information Matrix.

        Args:
            sample_size (int): The desired sample size.

        Raises:
            ValueError: If the model is not fitted.

        Returns:
            torch.Tensor: A tensor of shape (sample_size, n_params) of sampled model
                parameters as vectors.
        """
        check_is_fitted(self, "fim_")

        dist = torch.distributions.MultivariateNormal(
            parameters_to_vector(self.params.parameters()).detach(),
            cast(torch.Tensor, self.fim_).inverse(),
        )
        return dist.sample((sample_size,))

    @torch.no_grad()
    @validate_params(
        {
            "data": [ModelData],
            "u": [torch.Tensor],
            "n_samples": [Interval(Integral, 1, None, closed="left")],
            "double_monte_carlo": [bool],
        },
        prefer_skip_nested_validation=True,
    )
    def predict_y(
        self,
        data: ModelData,
        u: torch.Tensor,
        *,
        n_samples: int = 1000,
        double_monte_carlo: bool = False,
    ) -> torch.Tensor:
        r"""Function to predict longitudinal values.

        For every drawing of a random effect :math:`b`, this computes at the prediction
        times :math:`u` the values of the regression function given input data:

        .. math::
            h(u, b).

        The variable `u` is expected to be a matrix with the same number of rows as
        individuals, and the same number of columns as prediction times.

        If `double_monte_carlo` is True, then the prediction is computed using double
        Monte Carlo Rizopoulos-style (2011).

        Args:
            data (ModelData): The data to predict.
            u (torch.Tensor): The matrix containing prediction times.
            n_samples (int, optional): The number of samples to draw from the
                posterior. Defaults to 1000.
            double_monte_carlo (bool, optional): Whether to use double Monte Carlo.
                Defaults to False.

        Raises:
            ValueError: If `double_monte_carlo` is True and the model is not fitted.
            ValueError: If `u` contains inf or NaN values.
            ValueError: If `u` has incompatible shape.

        Returns:
            torch.Tensor: The predicted longitudinal values.
        """
        assert_all_finite(u, input_name="u")
        check_consistent_length(u, data)

        # Load and complete data
        data = CompleteModelData(data.x, data.t, data.y, data.trajectories, data.c)
        data.prepare(self.model_design, self.params)

        # Initialize variables
        y_pred: list[torch.Tensor] = []
        n_iter = ceil(n_samples / self.n_chains)

        if double_monte_carlo:
            init_param_vector = parameters_to_vector(self.params.parameters())
            vector_params = self._sample_vector_params(n_iter)

        # Initialize optimizer and MCMC
        sampler = self._init_mcmc(data)
        sampler.run(self.n_warmup)

        for i in trange(
            n_iter, desc="Predicting longitudinal values", disable=not self.verbose
        ):
            if double_monte_carlo:
                vector_to_parameters(vector_params[i], self.params.parameters())  # type: ignore

            y = self.model_design.regression_fn(u, sampler.psi)
            y_pred.extend(y[i] for i in range(y.size(0)))
            sampler.run(self.n_subsample)

        # Restore parameters
        if double_monte_carlo:
            vector_to_parameters(init_param_vector, self.params.parameters())  # type: ignore
        self._cache.clear()
        return torch.stack(y_pred[:n_samples])

    @torch.no_grad()
    @validate_params(
        {
            "data": [ModelData],
            "u": [torch.Tensor],
            "n_samples": [Interval(Integral, 1, None, closed="left")],
            "double_monte_carlo": [bool],
        },
        prefer_skip_nested_validation=True,
    )
    def predict_surv_logps(
        self,
        data: ModelData,
        u: torch.Tensor,
        *,
        n_samples: int = 1000,
        double_monte_carlo: bool = False,
    ) -> torch.Tensor:
        r"""Function to predict survival log probability values.

        For every drawing of a random effect :math:`b`, this computes at the prediction
        times :math:`u` the values of the log survival probabilities given input data
        and conditionally to survival up to time :math:`c`:

        .. math::
            \log \mathbb{P}(T^* \geq u \mid T^* > c) = -\int_c^u \lambda(t) \, dt.

        When multiple transitions are allowed, :math:`\lambda(t)` is a sum over all
            possible transitions, that is to say if an individual is in the state
            :math:`k` from time :math:`t_0`, this gives:

            .. math::
                -\int_c^u \sum_{k'} \lambda^{k' \mid k}(t \mid t_0) \, dt.

        Please note this makes use of the Chasles property in order to avoid the
        computation of two integrals and make computations more precise.

        The variable `u` is expected to be a matrix with the same number of rows as
        individuals, and the same number of columns as prediction times.

        If `double_monte_carlo` is True, then the prediction is computed using double
        Monte Carlo Rizopoulos-style (2011).

        Args:
            data (ModelData): The data to predict.
            u (torch.Tensor): The matrix containing prediction times.
            n_samples (int, optional): The number of samples to draw from the
                posterior. Defaults to 1000.
            double_monte_carlo (bool, optional): Whether to use double Monte Carlo.
                Defaults to False.

        Raises:
            ValueError: If `double_monte_carlo` is True and the model is not fitted.
            ValueError: If `u` contains inf or NaN values.
            ValueError: If `u` has incompatible shape.

        Returns:
            torch.Tensor: The predicted survival log probabilities.
        """
        assert_all_finite(u, input_name="u")
        check_consistent_length(u, data)

        # Load and complete data
        data = CompleteModelData(data.x, data.t, data.y, data.trajectories, data.c)
        data.prepare(self.model_design, self.params)

        # Initialize variables
        surv_logps_pred: list[torch.Tensor] = []
        n_iter = ceil(n_samples / self.n_chains)

        if double_monte_carlo:
            init_param_vector = parameters_to_vector(self.params.parameters())
            vector_params = self._sample_vector_params(n_iter)

        # Initialize optimizer and MCMC
        sampler = self._init_mcmc(data)
        sampler.run(self.n_warmup)

        for i in trange(
            n_iter,
            desc="Predicting survival log probabilities",
            disable=not self.verbose,
        ):
            if double_monte_carlo:
                vector_to_parameters(vector_params[i], self.params.parameters())  # type: ignore

            sample_data = SampleData(data.x, data.trajectories, sampler.psi, data.c)
            surv_logps = self.compute_surv_logps(sample_data, u)
            surv_logps_pred.extend(surv_logps[i] for i in range(surv_logps.size(0)))
            sampler.run(self.n_subsample)

        if double_monte_carlo:
            vector_to_parameters(init_param_vector, self.params.parameters())  # type: ignore
        self._cache.clear()
        return torch.stack(surv_logps_pred[:n_samples])

    @torch.no_grad()
    @validate_params(
        {
            "data": [ModelData],
            "c_max": [torch.Tensor],
            "max_length": [Interval(Integral, 1, None, closed="left")],
            "n_samples": [Interval(Integral, 1, None, closed="left")],
            "double_monte_carlo": [bool],
        },
        prefer_skip_nested_validation=True,
    )
    def predict_trajectories(
        self,
        data: ModelData,
        c_max: torch.Tensor,
        *,
        max_length: int = 10,
        n_samples: int = 1000,
        double_monte_carlo: bool = False,
    ) -> list[list[Trajectory]]:
        r"""Function to predict trajectories.

        For every drawing of a random effect :math:`b`, this simulates the trajectories
        up to time `c_max` with a maximum length of `max_length` to avoid infinite
        loops. This uses a variant of Gillepsie's algorithm.

        The variable `c_max` is expected to be a column vector with the same number of
        rows as individuals.

        If `double_monte_carlo` is True, then the prediction is computed using double
        Monte Carlo Rizopoulos-style (2011).

        Args:
            data (ModelData): The data to predict.
            c_max (torch.Tensor): The matrix containing prediction times.
            max_length (int, optional): The maximum length of the trajectories.
                Defaults to 10.
            n_samples (int, optional): The number of samples to draw from the
                posterior. Defaults to 1000.
            double_monte_carlo (bool, optional): Whether to use double Monte Carlo.
                Defaults to False.

        Raises:
            ValueError: If `double_monte_carlo` is True and the model is not fitted.
            ValueError: If `c_max` contains inf or NaN values.
            ValueError: If `c_max` has incompatible shape.

        Returns:
            torch.Tensor: The predicted trajectories.
        """
        assert_all_finite(c_max, input_name="c_max")
        check_consistent_length(c_max, data)

        data = CompleteModelData(data.x, data.t, data.y, data.trajectories, data.c)
        data.prepare(self.model_design, self.params)

        trajectories_pred: list[list[Trajectory]] = []
        n_iter = ceil(n_samples / self.n_chains)

        if double_monte_carlo:
            init_param_vector = parameters_to_vector(self.params.parameters())
            vector_params = self._sample_vector_params(n_iter)

        sampler = self._init_mcmc(data)
        sampler.run(self.n_warmup)

        for i in trange(
            n_iter,
            desc="Predicting trajectories",
            disable=not self.verbose,
        ):
            if double_monte_carlo:
                vector_to_parameters(vector_params[i], self.params.parameters())  # type: ignore

            for j in range(sampler.psi.size(0)):
                sample_data = SampleData(
                    data.x, data.trajectories, sampler.psi[j], data.c
                )
                trajectories_pred.append(
                    self.sample_trajectories(sample_data, c_max, max_length=max_length)
                )
            sampler.run(self.n_subsample)

        if double_monte_carlo:
            vector_to_parameters(init_param_vector, self.params.parameters())  # type: ignore
        self._cache.clear()
        return trajectories_pred[:n_samples]



from numbers import Real
from typing import cast

import torch
from sklearn.utils._param_validation import Interval, validate_params  # type: ignore
from torch import nn

from ..typedefs._defs import LOG_TWO_PI, ClockMethod, LogBaseHazardFn


def clock_forward(t0: torch.Tensor, t1: torch.Tensor) -> torch.Tensor:  # noqa: ARG001
    r"""Time transformation for clock forward method.

    This is the simple mapping:

    .. math::
        (t_0, t_1) \mapsto t_1.

    This type of mapping is particularly useful when the base risk does not depend on
    relative (sojourn) type, but rather on absolute time.

    Args:
        t0 (torch.Tensor): Past transition time.
        t1 (torch.Tensor): Current time

    Returns:
        torch.Tensor: Current time.

    Examples:
        >>> clock_forward(1, 2)
        2
    """
    return t1


def clock_reset(t0: torch.Tensor, t1: torch.Tensor) -> torch.Tensor:
    r"""Time transformation for clock reset method.

    This is the simple mapping:

    .. math::
        (t_0, t_1) \mapsto t_1 - t_0.

    This type of mapping is particularly useful when the base risk depends on
    relative (sojourn) type, but not on absolute time.

    Args:
        t0 (torch.Tensor): Past transition time.
        t1 (torch.Tensor): Current time

    Returns:
        torch.Tensor: Current time - Past transition time.

    Examples:
        >>> clock_reset(1, 2)
        1
    """
    return t1 - t0


class Exponential(LogBaseHazardFn):
    r"""Implements the Exponential base hazard.

    Exponential base hazard is time independent.

    It is given by the formula:

    .. math::
        \lambda(t) = \lambda.

    This returns the base hazard in log scale.$

    Attributes:
        log_lmda (nn.Parameter): The log rate factor.
    """

    log_lmda: nn.Parameter

    @validate_params(
        {
            "lmda": [Interval(Real, 0, None, closed="neither")],
        },
        prefer_skip_nested_validation=True,
    )
    def __init__(self, lmda: float):
        """Initializes the Exponential hazard.

        Args:
            lmda (float): The rate factor.
        """
        super().__init__()  # type: ignore

        lmda_ = torch.tensor(lmda)
        self.log_lmda = nn.Parameter(torch.log(lmda_))

    def forward(
        self,
        t0: torch.Tensor,  # noqa: ARG002
        t1: torch.Tensor,  # noqa: ARG002
    ) -> torch.Tensor:
        """Calls the Exponential base hazard.

        Args:
            t0 (torch.Tensor): Past transition time.
            t1 (torch.Tensor): Current time

        Returns:
            torch.Tensor: The computed base hazard in log scale.
        """
        return self.log_lmda

    @property
    def lmda(self) -> torch.Tensor:
        """Gets the rate factor.

        Returns:
            torch.Tensor: The rate factor.
        """
        return self.log_lmda.exp()


class Weibull(LogBaseHazardFn):
    r"""Implements the Weibull base hazard.

    Weibull base hazard is time dependent.

    It is given by the formula:

    .. math::
        \lambda(t) = \frac{k}{\lambda} \frac{t}{\lambda}^{k - 1}.

    This returns the base hazard in log scale.

    Attributes:
        clock_method (ClockMethod): The ClockMethod transformation.
        log_k (nn.Parameter): The log of the shape parameter.
        log_lmda (nn.Parameter): The log of the scale parameter.
    """

    clock_method: ClockMethod
    log_lmda: nn.Parameter
    log_k: nn.Parameter

    @validate_params(
        {
            "lmda": [Interval(Real, 0, None, closed="neither")],
            "k": [Interval(Real, 0, None, closed="neither")],
        },
        prefer_skip_nested_validation=True,
    )
    def __init__(self, lmda: float, k: float, clock_method: ClockMethod = clock_reset):
        """Initializes the Weibull base hazard.

        Args:
            lmda (float): The scale parameter.
            k (float): The shape parameter.
            clock_method (ClockMethod, optional): The ClockMethod transformation.
                Defaults to clock_reset.
        """
        super().__init__()  # type: ignore

        k_ = torch.tensor(k)
        lmda_ = torch.tensor(lmda)
        self.clock_method = clock_method
        self.log_k = nn.Parameter(torch.log(k_))
        self.log_lmda = nn.Parameter(torch.log(lmda_))

    def forward(self, t0: torch.Tensor, t1: torch.Tensor) -> torch.Tensor:
        """Calls the Weibull base hazard.

        Args:
            t0 (torch.Tensor): Past transition time.
            t1 (torch.Tensor): Current time

        Returns:
            torch.Tensor: The computed base hazard in log scale.
        """
        t = self.clock_method(t0, t1)
        log_t = torch.log(t)
        return self.log_k - self.log_lmda + (self.k - 1) * (log_t - self.log_lmda)

    @property
    def k(self) -> torch.Tensor:
        """Gets the shape parameter.

        Returns:
            torch.Tensor: The shape parameter.
        """
        return self.log_k.exp()

    @property
    def lmda(self) -> torch.Tensor:
        """Gets the scale parameter.

        Returns:
            torch.Tensor: The scale parameter.
        """
        return self.log_lmda.exp()


class Gompertz(LogBaseHazardFn):
    r"""Implements the Gompertz base hazard.

    Gompertz base hazard is time dependent.
    It is given by the formula:

    .. math::
        \lambda(t) = a \exp{bt}.

    This returns the base hazard in log scale.

    Attributes:
        b (nn.Parameter): The shape parameter.
        clock_method (ClockMethod): The ClockMethod transformation.
        log_a (nn.Parameter): The baseline hazard parameter.
    """

    b: nn.Parameter
    clock_method: ClockMethod
    log_a: nn.Parameter

    @validate_params(
        {
            "a": [Interval(Real, 0, None, closed="neither")],
            "b": [Interval(Real, None, None, closed="neither")],
        },
        prefer_skip_nested_validation=True,
    )
    def __init__(self, a: float, b: float, clock_method: ClockMethod = clock_reset):
        """Initializes the Gompertz base hazard.

        Args:
            a (float): The baseline hazard.
            b (float): The shape parameter.
            clock_method (ClockMethod, optional): The ClockMethod transformation.
                Defaults to clock_reset.
        """
        super().__init__()  # type: ignore

        a_ = torch.tensor(a)
        self.b = nn.Parameter(torch.tensor(b))
        self.clock_method = clock_method
        self.log_a = nn.Parameter(torch.log(a_))

    def __call__(self, t0: torch.Tensor, t1: torch.Tensor) -> torch.Tensor:
        """Calls the Gompertz base hazard.

        Args:
            t0 (torch.Tensor): Past transition time.
            t1 (torch.Tensor): Current time

        Returns:
            torch.Tensor: The computed base hazard in log scale.
        """
        t = self.clock_method(t0, t1)
        return self.log_a + self.b * t

    @property
    def a(self) -> torch.Tensor:
        """Gets the baseline hazard.

        Returns:
            torch.Tensor: The baseline hazard.
        """
        return self.log_a.exp()


class LogNormal(LogBaseHazardFn):
    r"""Implements the log normal base hazard.

    Log normal base hazard is time dependent.
    It is given by the formula :

    .. math::
        \lambda(t) = \frac{\phi\left( \frac{\log t - \mu}{\sigma} \right)}{t \sigma
        \, \Phi\left( -\frac{\log t - \mu}{\sigma} \right)},
        \quad t > 0,

    where:

    .. math::
        \phi(z) = \frac{1}{\sqrt{2\pi}} e^{-z^2/2}, \quad
        \Phi(z) = \frac{1}{\sqrt{2\pi}} \int_{-\infty}^z e^{-t^2/2} \, dt.

    This returns the base hazard in log scale.

    Attributes:
        mu (nn.Parameter): The log time mean.
        clock_method (ClockMethod): The ClockMethod transformation.
        log_scale (nn.Parameter): The log of scale.
    """

    mu: nn.Parameter
    clock_method: ClockMethod
    log_scale: nn.Parameter

    @validate_params(
        {
            "mu": [Interval(Real, None, None, closed="neither")],
            "scale": [Interval(Real, 0, None, closed="neither")],
        },
        prefer_skip_nested_validation=True,
    )
    def __init__(
        self, mu: float, scale: float, clock_method: ClockMethod = clock_reset
    ):
        """Initializes the log normal base hazard.

        Args:
            mu (float): The log time mean.
            scale (float): The log time scale.
            clock_method (ClockMethod, optional): The ClockMethod transformation.
                Defaults to clock_reset.

        Returns:
            LogBaseHazardFn: Returns the log normal base hazard function.
        """
        super().__init__()  # type: ignore

        self.mu = nn.Parameter(torch.tensor(mu))
        scale_ = torch.tensor(scale)
        self.clock_method = clock_method
        self.log_scale = nn.Parameter(torch.log(scale_))

    def forward(self, t0: torch.Tensor, t1: torch.Tensor) -> torch.Tensor:
        """Calls the log normal base hazard.

        Args:
            t0 (torch.Tensor): Past transition time.
            t1 (torch.Tensor): Current time

        Returns:
            torch.Tensor: The computed base hazard in log scale.
        """
        t = self.clock_method(t0, t1)
        log_t = torch.log(t)
        z = (log_t - self.mu) / self.scale
        log_pdf = -log_t - self.log_scale - 0.5 * LOG_TWO_PI - 0.5 * z**2
        log_sf = cast(torch.Tensor, torch.special.log_ndtr(-z))  # type: ignore
        return log_pdf - log_sf

    @property
    def scale(self) -> torch.Tensor:
        """Gets the scale.

        Returns:
            torch.Tensor: The scale.
        """
        return self.log_scale.exp()



from bisect import bisect_left

import torch
from rich.console import Console, Group
from rich.panel import Panel
from rich.rule import Rule
from rich.table import Table
from rich.text import Text
from torch.distributions import Normal
from torch.nn.utils import parameters_to_vector

from ..model._base import MultiStateJointModel
from ..typedefs._defs import SIGNIFICANCE_CODES, SIGNIFICANCE_LEVELS


def summary(model: MultiStateJointModel):
    """Prints a summary of the fitted model.

    This function prints the p-values of the parameters as well as values and
    standard error. Also prints the log likelihood, AIC, BIC with lovely colors.

    Raises:
        ValueError: If the model is not fitted.
    """
    vector = parameters_to_vector(model.params.parameters())
    stderr = model.stderr
    zvalues = torch.abs(vector / stderr)
    pvalues = 2 * (1 - Normal(0, 1).cdf(zvalues))

    table = Table()
    table.add_column("Parameter name", justify="left")
    table.add_column("Value", justify="center")
    table.add_column("Standard Error", justify="center")
    table.add_column("z-value", justify="center")
    table.add_column("p-value", justify="center")
    table.add_column("Significance level", justify="center")

    i = 0
    for name, val in model.params.named_parameters():
        for j in range(val.numel()):
            code = SIGNIFICANCE_CODES[
                bisect_left(SIGNIFICANCE_LEVELS, pvalues[i].item())
            ]

            table.add_row(
                f"{name}[{j}]",
                f"{vector[i].item():.3f}",
                f"{stderr[i].item():.3f}",
                f"{zvalues[i].item():.3f}",
                f"{pvalues[i].item():.3f}",
                code,
            )
            i += 1

    criteria = Text(
        f"Log-likelihood: {model.loglik_:.3f}\n"
        f"AIC: {model.aic_:.3f}\n"
        f"BIC: {model.bic_:.3f}",
        style="bold cyan",
    )

    content = Group(table, Rule(style="dim"), criteria, Rule(style="dim"))
    panel = Panel(content, title="Model Summary", border_style="green", expand=False)

    Console().print(panel)



"""Type definitions for the jmstate package."""

from ._data import CompleteModelData, ModelData, ModelDesign, SampleData
from ._defs import (
    BucketData,
    ClockMethod,
    IndividualEffectsFn,
    LinkFn,
    LogBaseHazardFn,
    RegressionFn,
)
from ._params import CovParam, ModelParams

__all__ = [
    "BucketData",
    "ClockMethod",
    "CompleteModelData",
    "CovParam",
    "IndividualEffectsFn",
    "LinkFn",
    "LogBaseHazardFn",
    "ModelData",
    "ModelDesign",
    "ModelParams",
    "RegressionFn",
    "SampleData",
]



from collections.abc import Iterator
from numbers import Integral
from typing import Any, Self, cast

import torch
from sklearn.base import BaseEstimator  # type: ignore
from sklearn.utils._param_validation import (  # type: ignore
    Interval,  # type: ignore
    StrOptions,  # type: ignore
    validate_params,  # type: ignore
)
from sklearn.utils.validation import assert_all_finite  # type: ignore
from torch import nn

from ..utils._checks import check_matrix_dim
from ..utils._linalg import flat_from_log_cholesky, log_cholesky_from_flat


class UniqueParams(nn.Module):
    """`nn.Module` that has unique parameters."""

    def parameters(self, recurse: bool = True) -> Iterator[nn.Parameter]:
        """Return an iterator over the unique parameters.

        Args:
            recurse (bool, optional): Whether to recurse into submodules. Defaults to
                True.

        Returns:
            Iterator[nn.Parameter]: An iterator over the unique parameters.
        """
        seen: set[int] = set()
        for param in super().parameters(recurse):
            if (ptr := param.data_ptr()) not in seen:
                seen.add(ptr)
                yield param

    def named_parameters(
        self, prefix: str = "", recurse: bool = True, remove_duplicate: bool = True
    ) -> Iterator[tuple[str, nn.Parameter]]:
        """Return an iterator over the unique parameters.

        Args:
            prefix (str, optional): The prefix to prepend to the parameter names.
                Defaults to "".
            recurse (bool, optional): Whether to recurse into submodules. Defaults to
                True.
            remove_duplicate (bool, optional): Whether to remove duplicate parameters.
                Defaults to True.

        Returns:
            Iterator[nn.Parameter]: An iterator over the unique parameters.
        """
        seen: set[int] = set()
        for name, param in super().named_parameters(prefix, recurse, remove_duplicate):
            if (ptr := param.data_ptr()) not in seen:
                seen.add(ptr)
                yield name, param


class CovParam(BaseEstimator, nn.Module):
    r"""`nn.Module` containing covariance parameters.

    Note three types of covariance matrices parametrization are provided: scalar
    matrix; diagonal matrix; full matrix. Defaults to the full matrix parametrization.
    This is achieved through a log Cholesky parametrization of the inverse covariance
    matrix. Formally, consider :math:`P = \Sigma^{-1}` the precision matrix and let
    :math:`L` be the Cholesky factor with positive diagonal elements, the log Cholseky
    is given by:

    .. math::
        \tilde{L}_{ij} = L_{ij}, \, i > j,

    and:

    .. math::
        \tilde{L}_{ii} = \log L_{ii}.

    This is very numerically stable and fast, as it doesn't require inverting the
    matrix when computing quadratic forms. The log determinant is then equal to:

    .. math::

        \log \det P = 2 \operatorname{Tr}(\tilde{L}).

    You can use these methods by creating the appropriate `MatRepr` with methods of
    `ball`, `diag` or `full`.

    Additionnally, if your data has mixed missing values, do not use `full` matrix
    parametrization for the residuals, as is this case the components must be
    independent.

    Attributes:
        flat (torch.Tensor): The flat representation of the covariance matrix.
        dim (int): The dimension of the covariance matrix.
        method (str): The method used to parametrize the covariance matrix.
    """

    @classmethod
    @validate_params(
        {
            "V": [torch.Tensor],
            "method": [StrOptions({"full", "diag", "ball"})],
        },
        prefer_skip_nested_validation=True,
    )
    def from_cov(cls, V: torch.Tensor, method: str = "full") -> Self:
        r"""Gets representation from covariance matrix according to choice of method.

        Args:
            V (torch.Tensor): The square covariance matrix parameter.
            method (str, optional): The method, `full`, `diag` or `ball`. Defaults to
                `full`.

        Returns:
            Self: The usable representation.
        """
        L = cast(torch.Tensor, torch.linalg.cholesky(V.inverse()))  # type: ignore
        L.diagonal().log_()
        return cls(flat_from_log_cholesky(L, method), L.size(0), method)

    @validate_params(
        {
            "flat": [torch.Tensor],
            "dim": [Interval(Integral, 1, None, closed="left")],
            "method": [StrOptions({"full", "diag", "ball"})],
        },
        prefer_skip_nested_validation=True,
    )
    def __init__(self, flat: torch.Tensor, dim: int, method: str):
        """Initializes the `CovParam` object.

        Args:
            flat (torch.Tensor): The flat representation of the covariance matrix.
            dim (int): The dimension of the covariance matrix.
            method (str): The method used to parametrize the covariance matrix.

        Raises:
            ValueError: If the representation is invalid.
        """
        super().__init__()

        self.flat = nn.Parameter(flat)
        self.dim = dim
        self.method = method

        check_matrix_dim(self.flat, self.dim, self.method)

    @property
    def cov(self) -> torch.Tensor:
        """Gets the covariance matrix.

        Returns:
            torch.Tensor: The covariance matrix.
        """
        L = log_cholesky_from_flat(self.flat, self.dim, self.method)
        L.diagonal().exp_()
        return torch.cholesky_inverse(L)

    @property
    def _inv_cholesky_and_log_eigvals(self) -> tuple[torch.Tensor, torch.Tensor]:
        """Gets Cholesky factor as well as log eigvals.

        Returns:
            tuple[torch.Tensor, torch.Tensor]: Precision matrix and log eigvals.
        """
        L = log_cholesky_from_flat(self.flat, self.dim, self.method)
        log_eigvals = 2 * L.diag()
        L.diagonal().exp_()

        return L, log_eigvals


class ModelParams(BaseEstimator, UniqueParams):
    r"""`nn.Module` containing model parameters.

    Shared parameters are possible by assigning the exact same object to multiple fields
    so that the data pointer is the same. `self.parameters()` will not return duplicate
    parameters and is safe to use.

    Note three types of covariance matrices parametrization are provided: scalar
    matrix; diagonal matrix; full matrix. Defaults to the full matrix parametrization.
    This is achieved through a log Cholesky parametrization of the inverse covariance
    matrix. Formally, consider :math:`P = \Sigma^{-1}` the precision matrix and let
    :math:`L` be the Cholesky factor with positive diagonal elements, the log Cholseky
    is given by:

    .. math::
        \tilde{L}_{ij} = L_{ij}, \, i > j,

    and:

    .. math::
        \tilde{L}_{ii} = \log L_{ii}.

    This is very numerically stable and fast, as it doesn't require inverting the
    matrix when computing quadratic forms. The log determinant is then equal to:

    .. math::

        \log \det P = 2 \operatorname{Tr}(\tilde{L}).

    You can use these methods by creating the appropriate `MatRepr` with methods of
    `ball`, `diag` or `full`.

    Additionnally, if your data has mixed missing values, do not use `full` matrix
    parametrization for the residuals, as is this case the components must be
    independent.

    Attributes:
        gamma (torch.Tensor | None): The population level parameters.
        q (CovParam): The random effects precision matrix representation.
        r (CovParam): The residuals precision matrix representation.
        alphas (nn.ParameterDict): The link linear parameters.
        betas (nn.ParameterDict | None): The covariates parameters.
        extra (nn.ParameterDict | None): A dictionary of parameters that is
            passed in addition to other mandatory parameters.

    Examples:
        >>> # To create a model with simple parameters
        >>> init_params = ModelParams(
        ...     torch.zeros_like(gamma),
        ...     CovParam.from_cov(torch.eye(Q.size(0)), method="diag"),
        ...     CovParam.from_cov(torch.eye(R.size(0)), method="ball"),
        ...     {k: torch.zeros_like(v) for k, v in alphas.items()},
        ...     {k: torch.zeros_like(v) for k, v in betas.items()},
        ... )
        >>> # To create a model with shared parameters
        >>> alpha_shared = torch.zeros_like(list(alphas.values())[0])
        >>> init_params_shared = ModelParams(
        ...     torch.zeros_like(gamma),
        ...     CovParam.from_cov(torch.eye(Q.size(0)), method="diag"),
        ...     CovParam.from_cov(torch.eye(R.size(0)), method="ball"),
        ...     {k: alpha_shared for k, v in alphas.items()},
        ...     {k: torch.zeros_like(v) for k, v in betas.items()},
        ... )
    """

    gamma: torch.Tensor | None
    q: CovParam
    r: CovParam
    alphas: nn.ParameterDict
    betas: nn.ParameterDict | None
    extra: nn.ParameterDict | None

    @validate_params(
        {
            "gamma": [torch.Tensor, None],
            "q": [CovParam],
            "r": [CovParam],
            "alphas": [dict],
            "betas": [dict],
            "extra": [dict, None],
        },
        prefer_skip_nested_validation=True,
    )
    def __init__(
        self,
        gamma: torch.Tensor | None,
        q: CovParam,
        r: CovParam,
        alphas: dict[tuple[Any, Any], torch.Tensor],
        betas: dict[tuple[Any, Any], torch.Tensor] | None,
        *,
        extra: dict[str, torch.Tensor] | None = None,
    ):
        """Initializes the `ModelParams` object.

        Args:
            gamma (torch.Tensor | None): The population level parameters.
            q (CovParam): The random effects precision matrix representation.
            r (CovParam): The residuals precision matrix representation.
            alphas (dict[tuple[Any, Any], torch.Tensor]): The link linear parameters.
            betas (dict[tuple[Any, Any], torch.Tensor] | None): The covariates
                parameters.
            extra (dict[str, torch.Tensor] | None, optional): A dictionary of parameters
                that is passed in addition to other mandatory parameters. Defaults to
                None.

        Raises:
            ValueError: If any of the tensors contains inf or NaN values.
        """
        super().__init__()

        self.gamma = None if gamma is None else nn.Parameter(gamma)
        self.q = q
        self.r = r
        self.alphas = nn.ParameterDict({str(key): val for key, val in alphas.items()})
        self.betas = (
            None
            if betas is None
            else nn.ParameterDict({str(key): val for key, val in betas.items()})
        )
        self.extra = None if extra is None else nn.ParameterDict(extra)

        for key, val in self.named_parameters():
            assert_all_finite(val.detach(), input_name=key)

    def numel(self) -> int:
        """Return the number of unique parameters.

        Returns:
            int: The number of the (unique) parameters.
        """
        return sum(p.numel() for p in self.parameters())



from collections import OrderedDict
from collections.abc import Callable
from typing import Any


class Cache:
    """Cache data."""

    cache_limit: int | None
    keys: tuple[str, ...]
    cache: dict[str, OrderedDict[tuple[int, ...], Any]]  # type: ignore

    def __init__(self, cache_limit: int | None, keys: tuple[str, ...]):
        """Inits the cache.

        Args:
            cache_limit (int | None): The cache limit, if None, means infinite.
            keys (tuple[str]): The keys of the cache.

        Raises:
            ValueError: If cache_limit is not None nor positive.
        """
        self.cache_limit = cache_limit
        self.keys = keys
        self.cache: dict[str, OrderedDict[tuple[int, ...], Any]] = {
            key: OrderedDict() for key in keys
        }

    def get_cache(self, name: str, key: Any, missing: Callable[[], Any]) -> Any:
        """Gets the cache [name][key].

        Args:
            name (str): The name of the cache info.
            key (Any): The key inside the current info.
            missing (Callable[[], Any]): The fallback function.

        Returns:
            Any: The cached element.
        """
        cache = self.cache[name]
        if key in cache:
            cache.move_to_end(key)
        else:
            cache[key] = missing()

        if self.cache_limit is not None and len(cache) > self.cache_limit:
            cache.popitem(last=False)

        return cache[key]

    def clear(self):
        """Clears the cache."""
        self.cache = {key: OrderedDict() for key in self.keys}



from dataclasses import replace
from numbers import Integral
from typing import Any, cast

import numpy as np
import torch
from sklearn.utils._param_validation import Interval, validate_params  # type: ignore
from sklearn.utils.validation import (  #  type: ignore
    assert_all_finite,  #  type: ignore
    check_consistent_length,  #  type: ignore
)
from xxhash import xxh3_64_intdigest

from ..typedefs._data import CompleteModelData, ModelDesign, SampleData
from ..typedefs._defs import HAZARD_CACHE_KEYS, HazardInfo, Trajectory
from ..typedefs._params import ModelParams
from ..utils._cache import Cache
from ..utils._surv import build_possible_buckets, build_remaining_buckets


class HazardMixin:
    """Mixin class for hazard model computations."""

    model_design: ModelDesign
    params: ModelParams
    n_quad: int
    n_bisect: int
    cache_limit: int | None
    _std_nodes: torch.Tensor
    _std_weights: torch.Tensor
    _cache: Cache

    def __init__(
        self,
        n_quad: int,
        n_bisect: int,
        cache_limit: int | None,
        *args: Any,
        **kwargs: Any,
    ):
        """Initializes the hazard mixin.

        Args:
            n_quad (int): Number of quadrature nodes.
            n_bisect (int): The number of bisection steps.
            cache_limit (int | None): Max length of cache.
        """
        super().__init__(*args, **kwargs)

        self.n_quad = n_quad
        self.n_bisect = n_bisect
        self.cache_limit = cache_limit
        self._std_nodes, self._std_weights = self._legendre_quad(n_quad)
        self._cache = Cache(cache_limit, HAZARD_CACHE_KEYS)

    @staticmethod
    def _legendre_quad(n_quad: int) -> tuple[torch.Tensor, torch.Tensor]:
        """Get the Legendre quadrature nodes and weights.

        Args:
            n_quad (int): The number of quadrature points.

        Returns:
            tuple[torch.Tensor, torch.Tensor]: The nodes and weights.
        """
        nodes, weights = cast(
            tuple[
                np.ndarray[Any, np.dtype[np.float64]],
                np.ndarray[Any, np.dtype[np.float64]],
            ],
            np.polynomial.legendre.leggauss(n_quad),  # type: ignore
        )

        dtype = torch.get_default_dtype()
        std_nodes = torch.tensor(nodes, dtype=dtype).unsqueeze(0)
        std_weights = torch.tensor(weights, dtype=dtype)

        return std_nodes, std_weights

    def _log_hazard(self, hazard_info: HazardInfo) -> torch.Tensor:
        """Computes log hazard.

        Args:
            hazard_info (HazardInfo): All necessary information for computation.

        Returns:
            torch.Tensor: The computed log hazard.
        """
        t0, t1, x, psi, alpha, beta, log_base_hazard_fn, link_fn = hazard_info

        # Compute baseline hazard
        def _base_create():
            return log_base_hazard_fn(t0, t1)

        if self.cache_limit != 0:
            try:
                key = (
                    *hazard_info.log_base_hazard_fn.key,
                    xxh3_64_intdigest(t0.detach().contiguous().numpy()),  # type: ignore
                    xxh3_64_intdigest(t1.detach().contiguous().numpy()),  # type: ignore
                )
                base = self._cache.get_cache("base", key, _base_create)
            except RuntimeError:
                base = _base_create()
        else:
            base = _base_create()

        # Compute time-varying effects
        mod = link_fn(t1, psi) @ alpha

        # Compute covariates effect
        if x is None or beta is None:
            return base + mod

        return base + mod + x @ beta.unsqueeze(-1)

    def _cum_hazard(
        self, hazard_info: HazardInfo, c: torch.Tensor | None
    ) -> torch.Tensor:
        """Computes cumulative hazard.

        Args:
            hazard_info (HazardInfo): All necessary information for computation.
            c (torch.Tensor | None): Integration start or censoring time, t0 if None.

        Returns:
            torch.Tensor: The computed cumulative hazard.
        """
        t0, t1, *_ = hazard_info
        c = t0 if c is None else c

        # Transform to quadrature interval
        def _half_create():
            return 0.5 * (t1 - c)

        def _quad_c_create():
            return (
                (c + t1)
                .reshape(-1, 1)
                .addmm(half.reshape(-1, 1), self._std_nodes, beta=0.5)
                .reshape(t1.size(0), -1)
            )

        if self.cache_limit != 0:
            try:
                key = (
                    xxh3_64_intdigest(c.detach().contiguous().numpy()),  # type: ignore
                    xxh3_64_intdigest(t1.detach().contiguous().numpy()),  # type: ignore
                )
                half = self._cache.get_cache("half", key, _half_create)
                quad_c = self._cache.get_cache("quad_c", key, _quad_c_create)
            except RuntimeError:
                half = _half_create()
                quad_c = _quad_c_create()
        else:
            half = _half_create()
            quad_c = _quad_c_create()

        # Compute hazard at quadrature points
        vals = self._log_hazard(hazard_info._replace(t1=quad_c))
        vals.clamp_(max=50.0).exp_()

        return half.reshape(t1.shape) * (
            vals.reshape(*vals.shape[:-1], -1, self._std_weights.size(-1))
            @ self._std_weights
        )

    def _log_and_cum_hazard(
        self,
        hazard_info: HazardInfo,
        enable_cache: bool,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Computes both log and cumulative hazard.

        Args:
            hazard_info (HazardInfo): All necessary information for computation.
            enable_cache (bool): Enables caching.

        Returns:
            tuple[torch.Tensor, torch.Tensor]: The log and cum hazard.
        """
        t0, t1, *_ = hazard_info

        # Transform to quadrature interval
        def _half_create():
            return 0.5 * (t1 - t0)

        def _quad_lc_create():
            return torch.cat(
                [t1, (t0 + t1).addmm(half, self._std_nodes, beta=0.5)], dim=-1
            )

        if enable_cache:
            try:
                key = (
                    xxh3_64_intdigest(t0.detach().contiguous().numpy()),  # type: ignore
                    xxh3_64_intdigest(t1.detach().contiguous().numpy()),  # type: ignore
                )
                half = self._cache.get_cache("half", key, _half_create)
                quad_lc = self._cache.get_cache("quad_c", key, _quad_lc_create)
            except RuntimeError:
                half = _half_create()
                quad_lc = _quad_lc_create()
        else:
            half = _half_create()
            quad_lc = _quad_lc_create()

        # Compute log hazard at all points
        all_vals = self._log_hazard(hazard_info._replace(t1=quad_lc))
        all_vals[..., 1:].clamp_(max=50.0).exp_()

        return all_vals[..., 0], half.reshape(-1) * (
            all_vals[..., 1:] @ self._std_weights
        )

    def _sample_transition(
        self,
        hazard_info: HazardInfo,
        c: torch.Tensor | None,
    ) -> torch.Tensor:
        """Sample survival times using inverse transform sampling.

        Args:
            hazard_info (HazardInfo): All necessary information for computation.
            c (torch.Tensor | None): Conditionning time.

        Returns:
            torch.Tensor: The computed pre transition times.
        """
        t0, t1, *_ = hazard_info

        # Initialize for bisection search
        t_left, t_right = (
            t0.clone(),
            torch.nextafter(t1, torch.tensor(torch.inf)),
        )

        # Generate exponential random variables
        target = -torch.log(torch.rand_like(t_left))

        # Bisection search for survival times
        for _ in range(self.n_bisect):
            t_mid = 0.5 * (t_left + t_right)

            surv_logps = self._cum_hazard(hazard_info._replace(t1=t_mid), c)

            # Update search bounds
            accept_mask = surv_logps.reshape(target.shape) < target
            torch.where(accept_mask, t_mid, t_left, out=t_left)
            torch.where(accept_mask, t_right, t_mid, out=t_right)

        return t_right

    def _sample_trajectory_step(
        self, sample_data: SampleData, c_max: torch.Tensor
    ) -> bool:
        """Appends the next simulated transition.

        Args:
            sample_data (SampleData): Sampling data
            c_max (torch.Tensor): Max sampling time.

        Returns:
            bool: If the sampling is done.
        """
        x = sample_data.x
        psi = sample_data.psi
        c = sample_data.c

        # Get buckets from last states
        current_buckets = build_possible_buckets(
            sample_data.trajectories, c_max, tuple(self.model_design.surv_map.keys())
        )

        if not current_buckets:
            return True

        # Initialize candidate transition times
        n_transitions = len(current_buckets)
        t_candidates = torch.full((len(sample_data), n_transitions), torch.inf)

        for j, (key, bucket) in enumerate(current_buckets.items()):
            idxs, t0, t1 = bucket

            hazard_info = HazardInfo(
                t0,
                t1,
                None if x is None else x.index_select(-2, idxs),
                psi.index_select(-2, idxs),
                self.params.alphas[str(key)],
                None if self.params.betas is None else self.params.betas[str(key)],
                *self.model_design.surv_map[key],
            )

            # Sample transition times
            t_sample = self._sample_transition(
                hazard_info,
                None if c is None else c.index_select(-2, idxs),
            )
            t_candidates[idxs, j] = t_sample.reshape(-1)

        # Find earliest transition
        min_times, argmin_idxs = torch.min(t_candidates, dim=1)
        bucket_keys = list(current_buckets.keys())

        for i, (time, arg_idx) in enumerate(zip(min_times, argmin_idxs, strict=True)):
            if torch.isfinite(time):
                sample_data.trajectories[i].append(
                    (time.item(), bucket_keys[int(arg_idx)][1])
                )

        return False

    @validate_params(
        {
            "sample_data": [SampleData],
            "c_max": [torch.Tensor],
            "max_length": [Interval(Integral, 1, None, closed="left")],
        },
        prefer_skip_nested_validation=True,
    )
    def sample_trajectories(
        self,
        sample_data: SampleData,
        c_max: torch.Tensor,
        *,
        max_length: int = 10,
    ) -> list[Trajectory]:
        """Sample trajectories from the joint model.

        The sampling is done usign a bisection algorithm by inversing the log cdf of the
        transitions inside a Gillespie-like algorithm.

        Args:
            sample_data (SampleData): Prediction data.
            c_max (torch.Tensor): The maximum trajectory censoring times.
            max_length (int, optional): Maximum iterations or sampling. Defaults to 10.

        Raises:
            ValueError: If `c_max` contains inf or NaN values.
            ValueError: If `c_max` has incorrect shape.

        Returns:
            list[Trajectory]: The sampled trajectories.
        """
        assert_all_finite(c_max, input_name="c_max")
        check_consistent_length(c_max, sample_data)
        sample_data_copied = replace(sample_data)

        # Sample future transitions iteratively
        for _ in range(max_length):
            if self._sample_trajectory_step(sample_data_copied, c_max):
                break
            object.__setattr__(sample_data_copied, "c", None)

        return [
            trajectory if trajectory[-1][0] <= c_max[i] else trajectory[:-1]
            for i, trajectory in enumerate(sample_data_copied.trajectories)
        ]

    @validate_params(
        {
            "sample_data": [SampleData],
            "u": [torch.Tensor],
        },
        prefer_skip_nested_validation=True,
    )
    def compute_surv_logps(
        self, sample_data: SampleData, u: torch.Tensor
    ) -> torch.Tensor:
        r"""Computes log probabilites of remaining event free up to time u.

        A censoring time may also be given. With known individual effects, this computes
        at the times :math:`u` the values of the log survival probabilities given input
        data conditionally to survival up to time :math:`c`:

        .. math::
            \log \mathbb{P}(T^* \geq u \mid T^* > c) = -\int_c^u \lambda(t) \, dt.

        When multiple transitions are allowed, :math:`\lambda(t)` is a sum over all
        possible transitions, that is to say if an individual is in the state :math:`k`
        from time :math:`t_0`, this gives:

        .. math::
            -\int_c^u \sum_{k'} \lambda^{k' \mid k}(t \mid t_0) \, dt.

        Please note this makes use of the Chasles property in order to avoid the
        computation of two integrals and make computations more precise.

        The variable `u` is expected to be a matrix with the same number of rows as
        individuals, and the same number of columns as prediction times.

        Args:
            sample_data (SampleData): The data on which to compute the probabilities.
            u (torch.Tensor): The time at which to evaluate the probabilities.

        Raises:
            ValueError: If `u` contains inf or NaN values.
            ValueError: If `u` has incorrect shape.

        Returns:
            torch.Tensor: The computed survival log probabilities.
        """
        assert_all_finite(u, input_name="u")
        check_consistent_length(u, sample_data)

        x = sample_data.x
        psi = sample_data.psi
        c = sample_data.c

        # Get buckets from last states
        buckets = build_remaining_buckets(
            sample_data.trajectories, tuple(self.model_design.surv_map.keys())
        )

        # Compute the log probabilities summing over transitions
        nlogps = torch.zeros(*psi.shape[:-1], u.size(1))
        for key, bucket in buckets.items():
            idxs, t0 = bucket

            hazard_info = HazardInfo(
                t0,
                u.index_select(0, idxs),
                None if x is None else x.index_select(-2, idxs),
                psi.index_select(-2, idxs),
                self.params.alphas[str(key)],
                None if self.params.betas is None else self.params.betas[str(key)],
                *self.model_design.surv_map[key],
            )

            # Compute negative log survival
            alts_logliks = self._cum_hazard(
                hazard_info, None if c is None else c.index_select(0, idxs)
            )
            nlogps.index_add_(-2, idxs, alts_logliks)

        return -nlogps.clamp(min=0.0)

    def _hazard_logliks(
        self, data: CompleteModelData, psi: torch.Tensor
    ) -> torch.Tensor:
        """Computes the hazard log likelihoods.

        Args:
            data (CompleteModelData): Dataset on which likelihood is computed.
            psi (torch.Tensor): A matrix of individual parameters.

        Returns:
            torch.Tensor: The computed log likelihoods.
        """
        logliks = torch.zeros(psi.shape[:-1])

        for key, bucket in data.buckets.items():
            idxs, t0, t1, obs = bucket

            hazard_info = HazardInfo(
                t0,
                t1,
                None if data.x is None else data.x.index_select(-2, idxs),
                psi.index_select(-2, idxs),
                self.params.alphas[str(key)],
                None if self.params.betas is None else self.params.betas[str(key)],
                *self.model_design.surv_map[key],
            )

            obs_logliks, alts_logliks = self._log_and_cum_hazard(
                hazard_info,
                self.cache_limit != 0,
            )
            vals = (-alts_logliks).addcmul(obs, obs_logliks)
            logliks.index_add_(-1, idxs, vals)

        return logliks



from abc import ABC, abstractmethod
from typing import Final, NamedTuple, Protocol, TypeAlias

import torch
from torch import nn
from xxhash import xxh3_64_intdigest

# Type Aliases
Trajectory: TypeAlias = list[tuple[float, str]]


# Protocols
class IndividualEffectsFn(Protocol):
    """The individual effects function protocol.

    Calls the individual effects function.

    It must be able to yield 2D or 3D tensors given inputs of `gamma` (population
    parameters), `x` (covariates matrix), and `b` (random effects). Note `b` is
    either 2D or 3D.

    Args:
        gamma (torch.Tensor | None): The population parameters.
        x (torch.Tensor | None): The fixed covariates matrix.
        b (torch.Tensor): The random effects.

    Returns:
        torch.Tensor: The individual parameters `psi`.

    Examples:
        >>> individual_effects_fn = lambda gamma, x, b: gamma + b
    """

    def __call__(
        self, gamma: torch.Tensor | None, x: torch.Tensor | None, b: torch.Tensor
    ) -> torch.Tensor: ...


class RegressionFn(Protocol):
    """The regression function protocol.

    It must be able to yield 3D and 4D tensors given 1D or 2D time inputs, as well
        as `psi` input of order 2 or 3. This is not very restrictive, but requires to be
        careful. The last dimension is the dimension of the response variable; second
        last is the repeated measurements; third last is individual based; possible
        fourth last is for parallelization of the MCMC sampler.

        It is identical to LinkFn.

    Args:
        t (torch.Tensor): The evaluation times.
        psi (torch.Tensor): The individual parameters.

    Returns:
        torch.Tensor: The response variable values.

    Examples:
        >>> def sigmoid(t: torch.Tensor, psi: torch.Tensor):
        ...     scale, offset, slope = psi.chunk(3, dim=-1)
        ...     # Fully broadcasted
        ...     return (scale * torch.sigmoid((t - offset) / slope)).unsqueeze(-1)
        >>> regression_fn = sigmoid
    """

    def __call__(self, t: torch.Tensor, psi: torch.Tensor) -> torch.Tensor: ...


class LinkFn(Protocol):
    """The link function protocol.

    It must be able to yield 3D and 4D tensors given 1D or 2D time inputs, as well
        as `psi` input of order 2 or 3. This is not very restrictive, but requires to be
        careful. The last dimension is the dimension of the response variable; second
        last is the repeated measurements; third last is individual based; possible
        fourth last is for parallelization of the MCMC sampler.

        It is identical to RegressionFn.

    Args:
        t (torch.Tensor): The evaluation times.
        psi (torch.Tensor): The individual parameters.

    Returns:
        torch.Tensor: The response variable values.

    Examples:
        >>> def sigmoid(t: torch.Tensor, psi: torch.Tensor):
        ...     scale, offset, slope = psi.chunk(3, dim=-1)
        ...     # Fully broadcasted
        ...     return (scale * torch.sigmoid((t - offset) / slope)).unsqueeze(-1)
        >>> link_fn = sigmoid
    """

    def __call__(self, t: torch.Tensor, psi: torch.Tensor) -> torch.Tensor: ...


class ClockMethod(Protocol):
    r"""The clock method protocol.

    This protocol is useful to differentiate between two natural mappings when dealing
    with base hazards:

    .. math::
        (t_0, t_1) \mapsto \begin{cases} t_1 - t_0 \text{(clock reset)}, \\ t_1
        \text{(clock forward)} \end{cases}.
    """

    def __call__(self, t0: torch.Tensor, t1: torch.Tensor) -> torch.Tensor: ...


# Abstract classes
class LogBaseHazardFn(nn.Module, ABC):
    """The log base hazard base class.

    This is not a protocol because caching is done, and therefore a key is required.
    Making this a `nn.Module`, one can check the value of the parameters and store their
    hashed values at the same time as the

    Note that the log base hazard function is in log scale, and expects a former
    transition time column vector `t0` as well as other times points at which the log
    base hazard is to be computed. `t1` is a matrix with the same number of rows as
    `t0`.

    Implement a `forward` and do not forget the `super().__init__()` when declaring your
    own class.

    Pass the parameters you want to optimize in the `ModelParams.extra` attribute as a
    list. If you do not want them to be optimized, then by default they do not require
    gradients for the given implementations.
    """

    @property
    def key(self) -> tuple[int, ...]:
        """Returns a hash containing the class type and parameters if there are any.

        Returns:
            tuple[int, ...]: A key used in caching operations.
        """
        ident = id(self)
        parameters = list(self.parameters())
        if parameters == []:
            return (ident,)

        params_flat = nn.utils.parameters_to_vector(self.parameters())
        return (ident, xxh3_64_intdigest(params_flat.detach().numpy()))  # type: ignore

    @abstractmethod
    def forward(self, t0: torch.Tensor, t1: torch.Tensor) -> torch.Tensor: ...


# Named tuples
class BucketData(NamedTuple):
    """A simple `NamedTuple` containing transition information.

    Attributes:
        idxs (torch.Tensor): The individual indices.
        t0 (torch.Tensor): A column vector of previous transition times.
        t1 (torch.Tensor): A column vecotr of next transition times.
    """

    idxs: torch.Tensor
    t0: torch.Tensor
    t1: torch.Tensor


class HazardInfo(NamedTuple):
    """A simple internal `NamedTuple` required for hazard computation.

    Attributes:
        t0 (torch.Tensor): A column vector of previous transition times.
        t1 (torch.Tensor): A matrix of next transition times.
        x (torch.Tensor | None): The covariates.
        psi (torch.Tensor): The individual parameters.
        alpha (torch.Tensor): The baseline hazard parameters.
        beta (torch.Tensor | None): The regression parameters.
        log_base_hazard_fn (LogBaseHazardFn): The log base hazard function.
        link_fn (LinkFn): The link function.
    """

    t0: torch.Tensor
    t1: torch.Tensor
    x: torch.Tensor | None
    psi: torch.Tensor
    alpha: torch.Tensor
    beta: torch.Tensor | None
    log_base_hazard_fn: LogBaseHazardFn
    link_fn: LinkFn


# Constants
LOG_TWO_PI: Final[torch.Tensor] = torch.log(torch.tensor(2.0 * torch.pi))
HAZARD_CACHE_KEYS: Final[tuple[str, ...]] = ("base", "half", "quad_c", "quad_lc")
SIGNIFICANCE_LEVELS: Final[tuple[float, ...]] = (
    0.001,
    0.01,
    0.05,
    0.1,
    float("inf"),
)
SIGNIFICANCE_CODES: Final[tuple[str, ...]] = (
    "[red3]***[/]",
    "[orange3]**[/]",
    "[yellow3]*[/]",
    ".",
    "",
)



from typing import cast

import torch


def identity(
    gamma: torch.Tensor | None,  # noqa: ARG001
    x: torch.Tensor | None,  # noqa: ARG001
    b: torch.Tensor,
) -> torch.Tensor:
    r"""The standard identity transformation.

    It is useful when only random effects are considered as individual parameters.

    This is simply the mapping:

    .. math::
        (\gamma, x, b) \mapsto b.

    Args:
        gamma (torch.Tensor | None): The population parameters.
        x (torch.Tensor | None): The (fixed) covariates.
        b (torch.Tensor): The random effects.

    Returns:
        torch.Tensor: The identity b
    """
    return b


def gamma_x_plus_b(
    gamma: torch.Tensor | None, x: torch.Tensor | None, b: torch.Tensor
) -> torch.Tensor:
    r"""The standard linear transformation.

    It is useful when a linear combination of covariates and random effects are
    considered as individual parameters.

    This is simply the mapping:

    .. math::
        (\gamma, x, b) \mapsto \gamma x + b

    Args:
        gamma (torch.Tensor | None): The population parameters. It must not be None.
        x (torch.Tensor | None): The (fixed) covariates. It must not be None.
        b (torch.Tensor): The random effects.

    Returns:
        torch.Tensor: The transformation x @ gamma + b
    """
    return cast(torch.Tensor, x) @ cast(torch.Tensor, gamma) + b


def gamma_plus_b(
    gamma: torch.Tensor | None,
    x: torch.Tensor | None,  # noqa: ARG001
    b: torch.Tensor,
) -> torch.Tensor:
    r"""The standard linear transformation.

    It is useful when a linear combination of population parameters and random effects
    are considered as individual parameters.

    This is simply the mapping:

    .. math::
        (\gamma, x, b) \mapsto \gamma + b

    Args:
        gamma (torch.Tensor | None): The population parameters. It must not be None.
        x (torch.Tensor | None): The (fixed) covariates.
        b (torch.Tensor): The random effects.

    Returns:
        torch.Tensor: The transformation gamma + b
    """
    return cast(torch.Tensor, gamma) + b



from collections.abc import Callable
from math import ceil
from typing import Any, Self, cast
from warnings import warn

import torch
from sklearn.utils._param_validation import validate_params  # type: ignore
from torch.func import functional_call, jacfwd  # type: ignore
from torch.nn.utils import parameters_to_vector
from tqdm import trange

from ..typedefs._data import CompleteModelData, ModelData, ModelDesign
from ..typedefs._params import ModelParams, UniqueParams
from ..utils._cache import Cache
from ._hazard import HazardMixin
from ._longitudinal import LongitudinalMixin
from ._prior import PriorMixin
from ._sampler import MCMCMixin, MetropolisHastingsSampler


class FitMixin(PriorMixin, LongitudinalMixin, HazardMixin, MCMCMixin, UniqueParams):
    """Mixin for fitting the model."""

    model_design: ModelDesign
    params: ModelParams
    optimizer: torch.optim.Optimizer | None
    n_warmup: int
    n_subsample: int
    max_iter_fit: int
    tol: float
    window_size: int
    n_samples_summary: int
    verbose: bool
    vector_params_history_: list[torch.Tensor]
    fim_: torch.Tensor | None
    loglik_: float | None
    aic_: float | None
    bic_: float | None
    _cache: Cache

    def __init__(
        self,
        optimizer: torch.optim.Optimizer | None,
        max_iter_fit: int,
        tol: float,
        window_size: int,
        n_samples_summary: int,
        *args: Any,
        **kwargs: Any,
    ):
        """Initializes the fit parameters.

        Args:
            optimizer (torch.optim.Optimizer): The optimizer.
            max_iter_fit (int): The maximum number of iterations for fitting.
            tol (float): The tolerance for the convergence.
            window_size (int): The window size for the convergence.
            n_samples_summary (int): The number of samples used to compute Fisher
                Information Matrix and model selection criteria.
        """
        super().__init__(*args, **kwargs)

        self.optimizer = optimizer
        self.max_iter_fit = max_iter_fit
        self.tol = tol
        self.window_size = window_size
        self.n_samples_summary = n_samples_summary

    def forward(self, data: CompleteModelData, b: torch.Tensor) -> torch.Tensor:
        """Computes the mean log pdfs for the random effects `b`.

        This is used to compute the Fisher Information Matrix, as
        `torch.func.functional_call` requires the function to implement a `forward`
        method that takes the parameters as the first argument.

        Args:
            data (CompleteModelData): Dataset on which likelihood is computed.
            b (torch.Tensor): The random effects.

        Returns:
           torch.Tensor: The log pdfs.
        """
        logpdfs, _ = self._logpdfs_aux_fn(data, b)
        return logpdfs if logpdfs.ndim == 1 else logpdfs.mean(dim=0)

    def _step(
        self,
        sampler: MetropolisHastingsSampler,
        data: CompleteModelData,
    ):
        """Performs a step of the optimizer.

        Args:
            sampler (MetropolisHastingsSampler): The sampler.
            data (CompleteModelData): The data.

        Raises:
            ValueError: If the optimizer is not initialized.
        """
        if self.optimizer is None:
            raise ValueError("Optimizer is not initialized.")

        def closure():
            self.optimizer.zero_grad()  # type: ignore
            logpdfs, _ = self._logpdfs_aux_fn(data, sampler.b)
            loss = -logpdfs.mean()
            loss.backward()  # type: ignore
            return loss.item()

        self.optimizer.step(closure)

        # Restore logpdfs and aux
        sampler.logpdfs, sampler.psi = sampler.logpdfs_aux_fn(sampler.b)

    def _is_converged(self) -> bool:
        """Checks if the optimizer has converged.

        This is based on a linear regression of the parameters with the current
        number of iterations. It R2 is below a threshold, the optimizer is
        considered to have converged.

        Args:
            optimizer (torch.optim.Adam): The optimizer.

        Returns:
            bool: True if the optimizer has converged, False otherwise.
        """

        def r2(Y: torch.Tensor) -> torch.Tensor:
            n = Y.size(0)
            i = torch.arange(n, dtype=torch.get_default_dtype())
            i_centered = i - (n - 1) / 2
            y_centered = Y - Y.mean(dim=0)
            num = (i_centered @ y_centered) ** 2
            den = i_centered.pow(2).sum() * y_centered.pow(2).sum(dim=0)
            return (num / den).nan_to_num()

        if len(self.vector_params_history_) < self.window_size:
            return False

        Y = torch.stack(self.vector_params_history_[-self.window_size :])
        return r2(Y).max().item() < self.tol

    def _init_jac(
        self, data: CompleteModelData
    ) -> tuple[
        torch.Tensor, Callable[[dict[str, torch.Tensor], torch.Tensor], torch.Tensor]
    ]:
        """Initializes the Jacobian matrix.

        Args:
            data (CompleteModelData): The complete model data.

        Returns:
            tuple[torch.Tensor, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]]:
                The Jacobian matrix and the Jacobian function.
        """

        @jacfwd  # type: ignore
        def _dict_jac_fn(params_dict: dict[str, torch.Tensor], b: torch.Tensor):
            return functional_call(self, params_dict, args=(data, b))

        def _jac_fn(params_dict: dict[str, torch.Tensor], b: torch.Tensor):
            return torch.cat(list(_dict_jac_fn(params_dict, b).values()), dim=-1)  # type: ignore

        return torch.zeros(len(data), self.params.numel()), cast(
            Callable[[dict[str, torch.Tensor], torch.Tensor], torch.Tensor], _jac_fn
        )

    def _update_jac(
        self,
        mjac: torch.Tensor,
        jac_fn: Callable[[dict[str, torch.Tensor], torch.Tensor], torch.Tensor],
        sampler: MetropolisHastingsSampler,
    ):
        """Updates the Jacobian matrix.

        Args:
            mjac (torch.Tensor): The mean Jacobian matrix.
            jac_fn (Callable[[torch.Tensor, torch.Tensor], torch.Tensor]): The Jacobian
                function.
            sampler (MetropolisHastingsSampler): The sampler.
        """
        params_dict = dict(self.named_parameters())
        mjac += jac_fn(params_dict, sampler.b).detach()  # type: ignore

    def _compute_fim(self, mjac: torch.Tensor) -> torch.Tensor:
        """Computes the Fisher Information Matrix.

        Args:
            mjac (torch.Tensor): The mean Jacobian matrix.

        Returns:
            torch.Tensor: The Fisher Information Matrix.
        """
        mjac /= ceil(self.n_samples_summary / self.n_chains)
        return mjac.T @ mjac

    def _init_criteria(self, data: CompleteModelData):
        """Initializes the criteria.

        Args:
            data (CompleteModelData): The complete model data.

        Returns:
            tuple[torch.Tensor, torch.Tensor, torch.Tensor]: The criteria.
        """
        logpdf = torch.tensor(0.0)
        mb = torch.zeros(len(data), self.params.q.dim)
        mb2 = torch.zeros(len(data), self.params.q.dim, self.params.q.dim)
        return logpdf, mb, mb2

    def _update_criteria(
        self,
        logpdf: torch.Tensor,
        mb: torch.Tensor,
        mb2: torch.Tensor,
        sampler: MetropolisHastingsSampler,
    ):
        """Updates the criteria.

        Args:
            logpdf (torch.Tensor): The logpdf.
            mb (torch.Tensor): The mean of b.
            mb2 (torch.Tensor): The mean of squared b.
            sampler (MetropolisHastingsSampler): The sampler.
        """
        logpdf += sampler.logpdfs.mean()
        mb += sampler.b.mean(dim=0)
        mb2 += torch.einsum("ijk,ijl->jkl", sampler.b, sampler.b)

    def _compute_criteria(
        self,
        logpdf: torch.Tensor,
        mb: torch.Tensor,
        mb2: torch.Tensor,
        fim: torch.Tensor,
    ) -> tuple[float, float, float]:
        """Computes the criteria.

        Args:
            logpdf (torch.Tensor): The logpdf.
            mb (torch.Tensor): The mean of b.
            mb2 (torch.Tensor): The mean of squared b.
            fim (torch.Tensor): The Fisher Information Matrix.

        Returns:
            tuple[float, float, float]: The criteria.
        """
        n_iter = ceil(self.n_samples_summary / self.n_chains)
        logpdf /= n_iter
        mb /= self.n_chains * n_iter
        mb2 /= self.n_chains * n_iter

        covs = mb2 - torch.einsum("ij,ik->ijk", mb, mb)
        entropy = 0.5 * (torch.logdet(covs) + self.params.q.dim).sum().item()
        loglik = logpdf.item() + entropy
        aic = -2 * loglik + 2 * self.params.numel()
        bic = -2 * loglik + torch.logdet(fim).item()
        return loglik, aic, bic

    @validate_params(
        {
            "data": [ModelData],
        },
        prefer_skip_nested_validation=True,
    )
    def fit(self, data: ModelData) -> Self:
        r"""Fits the model to the data.

        It leverages the Fisher identity and stochastic gradient algorithm coupled
        with a MCMC (Metropolis-Hastings) sampler. The Fisher identity states that

        .. math::
            \nabla_\theta \log \mathcal{L}(\theta ; x) = \mathbb{E}_{b \sim p(\cdot
            \mid x, \theta)} \left( \nabla_\theta \log \mathcal{L}(\theta ; x, b)
            \right).

        Many methods exist for computing the Fisher Information Matrix in latent
        variable models. In particular, this class leverages the expected Fisher
        Information Matrix using the identity

        .. math::
            \mathcal{I}_n(\theta) = \sum_{i=1}^n \mathbb{E}_{b \sim p(\cdot \mid x_i,
            \theta)} \left(\nabla \log \mathcal{L}(\theta ; x_i, b) \nabla \log
            \mathcal{L}(\theta ; x_i, b)^T \right).

        For more information, see ISSN 2824-7795.

        Args:
            data (ModelData): The data to fit the model to.

        Returns:
            Self: The fitted model.
        """
        data = CompleteModelData(data.x, data.t, data.y, data.trajectories, data.c)
        data.prepare(self.model_design, self.params)

        sampler = self._init_mcmc(data)
        sampler.run(self.n_warmup)

        # Main fitting loop
        for _ in trange(
            self.max_iter_fit, desc="Fitting joint model", disable=not self.verbose
        ):
            self._step(sampler, data)
            self.vector_params_history_.append(
                parameters_to_vector(self.params.parameters()).detach()
            )
            if self._is_converged():
                break
            sampler.run(self.n_subsample)

        if not self._is_converged():
            warn(
                "Model may not have converged in the specified number of iterations.",
                stacklevel=2,
            )

        # Initialize Jacobian matrix and criteria variables
        mjac, jac_fn = self._init_jac(data)
        logpdf, mb, mb2 = self._init_criteria(data)

        # FIM and Criteria loop
        n_iter = ceil(self.n_samples_summary / self.n_chains)
        for _ in trange(
            n_iter,
            desc="Computing FIM and Model Selection Criteria",
            disable=not self.verbose,
        ):
            self._update_jac(mjac, jac_fn, sampler)
            self._update_criteria(logpdf, mb, mb2, sampler)
            for _ in range(self.n_subsample):
                sampler.step()

        self.fim_ = self._compute_fim(mjac)
        self.loglik_, self.aic_, self.bic_ = self._compute_criteria(
            logpdf, mb, mb2, self.fim_
        )

        self._cache.clear()
        return self



import itertools
from array import array
from collections import defaultdict
from functools import lru_cache
from typing import Any

import torch

from ..typedefs._defs import BucketData, Trajectory


def build_buckets(
    trajectories: list[Trajectory],
) -> dict[tuple[Any, Any], BucketData]:
    """Builds buckets from trajectories for user convenience.

    This yeilds a `NamedTuple` containing transition information containing:
        idxs (torch.Tensor): The individual indices.
        t0 (torch.Tensor): A column vector of previous transition times.
        t1 (torch.Tensor): A column vector of next transition times.

    Args:
        trajectories (list[Trajectory]): The list of individual trajectories.

    Raises:
        TypeError: If the default dtype is not float32 or float64.

    Returns:
        dict[tuple[Any, Any], BucketData]: Transition keys with values (idxs, t0, t1).
    """
    dtype = torch.get_default_dtype()
    if dtype == torch.float32:
        typecode = "f"
    elif dtype == torch.float64:
        typecode = "d"
    else:
        raise TypeError(f"Unsupported default dtype: {dtype}")

    # Process each individual trajectory
    buckets: defaultdict[
        tuple[Any, Any], tuple[array[int], array[float], array[float]]
    ] = defaultdict(lambda: (array("q"), array(typecode), array(typecode)))

    for i, trajectory in enumerate(trajectories):
        for (t0, s0), (t1, s1) in itertools.pairwise(trajectory):
            idxs, t0s, t1s = buckets[(s0, s1)]
            idxs.append(i)
            t0s.append(t0)
            t1s.append(t1)

    result = {
        key: BucketData(
            torch.frombuffer(idxs, dtype=torch.int64),
            torch.frombuffer(t0s, dtype=dtype).reshape(-1, 1),
            torch.frombuffer(t1s, dtype=dtype).reshape(-1, 1),
        )
        for key, (idxs, t0s, t1s) in buckets.items()
    }

    return dict(sorted(result.items()))


@lru_cache
def _build_alt_map(
    surv_keys: tuple[tuple[Any, Any], ...],
) -> defaultdict[Any, tuple[tuple[Any, Any], ...]]:
    """Builds alternative state mapping as tuples in a defaultdict.

    Args:
        surv_keys (tuple[tuple[Any, Any], ...]): The survival keys.

    Returns:
        defaultdict[Any, tuple[tuple[Any, Any], ...]]: The alternative state map.
    """
    return defaultdict(
        lambda: (),
        {
            s0: tuple((k, v) for k, v in surv_keys if k == s0)
            for s0 in {s0 for s0, _ in surv_keys}
        },
    )


def build_all_buckets(
    trajectories: list[Trajectory],
    c: torch.Tensor,
    surv_keys: tuple[tuple[Any, Any], ...],
) -> dict[tuple[Any, Any], tuple[torch.Tensor, ...]]:
    """Build vectorizable bucket representation.

    Args:
        trajectories (list[Trajectory]): The trajectories.
        c (torch.Tensor): Censoring times.
        surv_keys (tuple[tuple[Any, Any], ...]): The survival keys.

    Raises:
        TypeError: If the default dtype is not float32 or float64.

    Returns:
        dict[tuple[Any, Any], tuple[torch.Tensor, ...]]: The vectorizable buckets
            representation.
    """
    alt_map = _build_alt_map(surv_keys)
    dtype = torch.get_default_dtype()
    if dtype == torch.float32:
        typecode = "f"
    elif dtype == torch.float64:
        typecode = "d"
    else:
        raise TypeError(f"Unsupported default dtype: {dtype}")

    # Initialize buckets
    buckets: defaultdict[
        tuple[Any, Any], tuple[array[int], array[float], array[float], array[bool]]
    ] = defaultdict(lambda: (array("q"), array(typecode), array(typecode), array("b")))

    # Process each individual trajectory
    for i, trajectory in enumerate(trajectories):
        for (t0, s0), (t1, s1) in itertools.pairwise(trajectory):
            for key in alt_map[s0]:
                idxs, t0s, t1s, obs = buckets[key]
                idxs.append(i)
                t0s.append(t0)
                t1s.append(t1)
                obs.append(key[1] == s1)

        (last_t, last_s), c_i = trajectory[-1], c[i].item()

        if last_t >= c_i:
            continue

        for key in alt_map[last_s]:
            idxs, t0s, t1s, obs = buckets[key]
            idxs.append(i)
            t0s.append(last_t)
            t1s.append(c_i)
            obs.append(False)

    return {
        key: (
            torch.frombuffer(idxs, dtype=torch.int64),
            torch.frombuffer(t0s, dtype=dtype).reshape(-1, 1),
            torch.frombuffer(t1s, dtype=dtype).reshape(-1, 1),
            torch.frombuffer(obs, dtype=torch.bool),
        )
        for key, (idxs, t0s, t1s, obs) in buckets.items()
    }


def build_possible_buckets(
    trajectories: list[Trajectory],
    c: torch.Tensor,
    surv_keys: tuple[tuple[Any, Any], ...],
) -> dict[tuple[Any, Any], tuple[torch.Tensor, ...]]:
    """Build possible bucket representation.

    Args:
        trajectories (list[Trajectory]): The trajectories.
        c (torch.Tensor): Censoring times.
        surv_keys (tuple[tuple[Any, Any], ...]): The survival keys.

    Raises:
        TypeError: If the default dtype is not float32 or float64.

    Returns:
        dict[tuple[Any, Any], tuple[torch.Tensor, ...]]: The possible buckets
            representation.
    """
    alt_map = _build_alt_map(surv_keys)
    dtype = torch.get_default_dtype()
    if dtype == torch.float32:
        typecode = "f"
    elif dtype == torch.float64:
        typecode = "d"
    else:
        raise TypeError(f"Unsupported default dtype: {dtype}")

    # Initialize buckets
    buckets: defaultdict[tuple[Any, Any], tuple[array[int], array[float]]] = (
        defaultdict(lambda: (array("q"), array(typecode)))
    )

    # Process each individual trajectory
    for i, trajectory in enumerate(trajectories):
        last_t, last_s = trajectory[-1]

        if last_t >= c[i].item():
            continue

        for key in alt_map[last_s]:
            idxs, t0s = buckets[key]
            idxs.append(i)
            t0s.append(last_t)

    return {
        key: (
            idxs_tensor := torch.frombuffer(idxs, dtype=torch.int64),
            torch.frombuffer(t0s, dtype=dtype).reshape(-1, 1),
            c.index_select(0, idxs_tensor),
        )
        for key, (idxs, t0s) in buckets.items()
    }


def build_remaining_buckets(
    trajectories: list[Trajectory],
    surv_keys: tuple[tuple[Any, Any], ...],
) -> dict[tuple[Any, Any], tuple[torch.Tensor, ...]]:
    """Build remaining bucket representation.

    Args:
        trajectories (list[Trajectory]): The trajectories.
        surv_keys (tuple[tuple[Any, Any], ...]): The survival keys.

    Raises:
        TypeError: If the default dtype is not float32 or float64.

    Returns:
        dict[tuple[Any, Any], tuple[torch.Tensor, ...]]: The remaining buckets
            representation.
    """
    alt_map = _build_alt_map(surv_keys)
    dtype = torch.get_default_dtype()
    if dtype == torch.float32:
        typecode = "f"
    elif dtype == torch.float64:
        typecode = "d"
    else:
        raise TypeError(f"Unsupported default dtype: {dtype}")

    # Initialize buckets
    buckets: defaultdict[tuple[Any, Any], tuple[array[int], array[float]]] = (
        defaultdict(lambda: (array("q"), array(typecode)))
    )

    # Process each individual trajectory
    for i, trajectory in enumerate(trajectories):
        last_t, last_s = trajectory[-1]
        for key in alt_map[last_s]:
            idxs, t1s = buckets[key]
            idxs.append(i)
            t1s.append(last_t)

    return {
        key: (
            torch.frombuffer(idxs, dtype=torch.int64),
            torch.frombuffer(t1s, dtype=dtype).reshape(-1, 1),
        )
        for key, (idxs, t1s) in buckets.items()
    }



from __future__ import annotations

import math
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import torch
from numpy import atleast_1d
from sklearn.utils._param_validation import validate_params  # type: ignore

from ..typedefs._params import ModelParams
from ..utils._checks import check_params_align

if TYPE_CHECKING:
    from ..model._base import MultiStateJointModel


@validate_params(
    {
        "true_params": [ModelParams, None],
        "figsize": [tuple],
        "show": [bool],
    },
    prefer_skip_nested_validation=True,
)
def plot_params_history(
    model: MultiStateJointModel,
    *,
    true_params: ModelParams | None = None,
    figsize: tuple[int, int] = (10, 8),
    show: bool = True,
):
    """Plot the history of the parameters.

    This function plots the history of the parameters in a grid of subplots.

    Args:
        model (MultiStateJointModel): The model to plot the parameters history of.
        true_params (ModelParams | None, optional): The real parameters. Defaults to
            None.
        figsize (tuple[int, int], optional): The figure size. Defaults to (10, 8).
        show (bool, optional): Whether to show the plot. Defaults to True.

    Raises:
        ValueError: If the model has less than two parameter history.
        ValueError: If the parameters do not have the same names.
        ValueError: If the parameters do not have the same shapes.
    """
    from ..model._base import MultiStateJointModel  # noqa: PLC0415

    validate_params(
        {
            "model": [MultiStateJointModel],
        },
        prefer_skip_nested_validation=True,
    )

    if len(model.vector_params_history_) <= 1:
        raise ValueError("Only one parameter history provided")

    # Get the names
    named_params_dict = dict(model.params.named_parameters())
    nsubplots = len(named_params_dict)
    ncols = math.ceil(math.sqrt(nsubplots))
    nrows = math.ceil(nsubplots / ncols)

    # Check alignment between model params and true params
    if true_params is not None:
        check_params_align(model.params, true_params)
        named_true_params_dict = dict(true_params.named_parameters())

    # Create subplots
    fig, axes = plt.subplots(nrows, ncols, figsize=figsize)  # type: ignore
    axes = atleast_1d(axes).flat

    params_matrix_history = torch.stack(model.vector_params_history_)
    i = 0
    for ax, (name, val) in zip(axes, named_params_dict.items(), strict=False):
        history = params_matrix_history[:, i : (i := i + val.numel())]
        lines = ax.plot(history, label=[f"{name}[{j}]" for j in range(val.numel())])

        if true_params is not None:
            for line, p in zip(
                lines,
                named_true_params_dict[name].detach(),  # type: ignore
                strict=True,
            ):
                ax.axhline(p, linestyle="--", color=line.get_color())

        ax.set(title=name, xlabel="Iteration", ylabel="Value")
        ax.legend()

    # Remove unused subplots
    for j in range(nsubplots, len(axes)):
        fig.delaxes(axes[j])

    plt.suptitle("Stochastic optimization of the parameters")  # type: ignore
    plt.tight_layout()
    if show:
        plt.show()  # type: ignore
